# -*- coding: utf-8 -*-

# --------------------------------------------------------------------------
# Copyright Commvault Systems, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# --------------------------------------------------------------------------

"""File for operating on a OneDrive Subclient.

OneDriveSubclient is the only class defined in this file.

OneDriveSubclient:    Derived class from CloudAppsSubclient Base class, representing a
OneDrive subclient, and to perform operations on that subclient

OneDriveSubclient:

    content()                                                           --  gets the content of the subclient

    groups()                                                            --  gets the groups associated with the subclient

    restore_out_of_place()                                              --  runs out-of-place restore for the subclient

    discover()                                                          --  runs user discovery on subclient

    add_AD_group()                                                      --  adds AD group to the subclient

    add_user()                                                          --  adds user to the subclient

    add_users_onedrive_for_business_client()                            --  Adds user to OneDrive for Business Client

    search_for_user()                                                   --  Searches for a specific user's details from
                                                                            discovered list

    disk_restore_onedrive_for_business_client()                         --  Runs disk restore of selected users for
                                                                            OneDrive for Business Client

    out_of_place_restore_onedrive_for_business_client()                 --  Runs out-of-place restore of selected users
                                                                            for OneDrive for Business Client

    in_place_restore_onedrive_syntex()          --  Runs in-place restore of selected users for Syntex OneDrive for Business Client

    in_place_restore_onedrive_for_business_client()               --  Runs in-place restore of selected users for OneDrive for Business Client

    _get_user_guids()                                                   --  Retrieve GUIDs for users specified

    _task_json_for_onedrive_backup()                                    --  Json for onedrive backup for selected users

    _association_users_json()                                           --  user association

    point_in_time_in_place_restore_onedrive_for_business_client()       -- Runs PIT in-place restore of selected users

    point_in_time_out_of_place_restore_onedrive_for_business_client()   -- Runs PIT out of place restore of selected users

    run_user_level_backup_onedrive_for_business_client()                --  Runs the backup for the users in users list

    _get_user_details()                                                 --   gets user details from discovery

    _get_group_details()                                                --   gets group details from discovery

    browse_for_content()                                                --  Returns the Onedrive client content i.e.
                                                                            users/group information that is discovered
                                                                            in auto discovery phase

    _set_properties_to_update_site_association()                        --   Updates the association properties of user

    update_users_association_properties()                               --   Updates the association properties of user

    manage_custom_category()                                            --   Adds or Edits Custom category in the office 365 app

    update_custom_categories_association_properties()                   --  Updates the association properties of custom category

    refresh_retention_stats()                                           --  refresh the retention stats for the client

    refresh_client_level_stats()                                        --  refresh the client level stats for the client

    get_client_level_stats()                                            --  Returns the client level stats for the client

"""

from __future__ import unicode_literals

import datetime

from ...exception import SDKException
import time
from ..casubclient import CloudAppsSubclient
from ...constants import AppIDAType
from .onedrive_constants import OneDriveConstants as constants
import re

class OneDriveSubclient(CloudAppsSubclient):
    """Derived class from CloudAppsSubclient Base class, representing a OneDrive subclient,
        and to perform operations on that subclient."""

    def _get_subclient_properties(self):
        """Gets the subclient  related properties of File System subclient.."""
        super(OneDriveSubclient, self)._get_subclient_properties()
        if 'content' in self._subclient_properties:
            self._content = self._subclient_properties['content']

        content = []
        group_list = []

        for account in self._content:
            temp_account = account["cloudconnectorContent"]["includeAccounts"]

            if temp_account['contentType'] == AppIDAType.CLOUD_APP.value:
                content_dict = {
                    'SMTPAddress': temp_account["contentName"].split(";")[0],
                    'display_name': temp_account["contentValue"]
                }

                content.append(content_dict)
            if temp_account['contentType'] == 135:
                group_list.append(temp_account["contentName"])
        self._ca_content = content
        self._ca_groups = group_list

    def _get_subclient_properties_json(self):
        """get the all subclient related properties of this subclient.

           Returns:
                dict - all subclient properties put inside a dict

        """

        return {'subClientProperties': self._subclient_properties}

    def _association_users_json(self, users_list):
        """
            Args:
                users_list (list) : list of SMTP addresses of users
            Returns:
                users_json(list): Required details of users to backup
        """
        users_json = []
        for user_smtp in users_list:
            user_details=self._get_user_details(user_smtp)
            user_info={
                      "user": {
                        "userGUID": user_details[0].get('user', {}).get('userGUID')
                      }
            }
            users_json.append(user_info)
        return users_json

    def _task_json_for_onedrive_backup(self, users_list, custom_groups_list=[]):
        """
        Json for onedrive backup for selected users

        Args:
                users_list (list) : list of SMTP addresses of users
                custom_groups_list (list) : list of custom category groups
        """
        groups, _ = self.browse_for_content(discovery_type=31)
        associated_users_json = self._association_users_json(users_list)

        associated_custom_groups_json = []
        if len(custom_groups_list) != 0:
            for group in custom_groups_list:
                group_info = {
                    "id": groups[group].get('id', None),
                    "name": group
                }
                associated_custom_groups_json.append(group_info)

        advanced_options_dict = {
            'cloudAppOptions': {
                'userAccounts': associated_users_json,
                'userGroups': associated_custom_groups_json
            }
        }

        selected_items=[]
        for user_smtp in users_list:
            details=self._get_user_details(user_smtp)
            item={
                "itemName": details[0].get('displayName'),
                "itemType": "User"
            }
            selected_items.append(item)

        for group in custom_groups_list:
            item={
                "itemName": group,
                "itemtype": "Custom category"
            }
            selected_items.append(item)

        common_options_dict={
            "jobMetadata": [
                {
                    "selectedItems": selected_items,
                    "jobOptionItems": [
                        {
                            "option": "Total running time",
                            "value": "Disabled"
                        }
                    ]
                }
            ]
        }

        task_json = self._backup_json(backup_level='INCREMENTAL',incremental_backup=False,incremental_level='BEFORE_SYNTH',advanced_options=advanced_options_dict,common_backup_options=common_options_dict)
        return task_json

    @property
    def content(self):
        """Returns the subclient content dict"""
        return self._ca_content

    @property
    def groups(self):
        """Returns the list of groups assigned to the subclient if any.
        Groups are assigned only if auto discovery is enabled for groups.

            Returns:

                list - list of groups associated with the subclient

        """
        return self._ca_groups

    @content.setter
    def content(self, subclient_content):
        """Creates the list of content JSON to pass to the API to add/update content of a
            Cloud Apps Subclient.

            Args:
                subclient_content (list)  --  list of the content to add to the subclient
                                              contains the account info for each user in list.

                                              example temp_content_dict={
                                                "cloudconnectorContent": {
                                                  "includeAccounts": {
                                                    "contentValue": Automation User,
                                                    "contentType": 134,
                                                    "contentName": automation_user@automationtenant.com
                                                     }
                                                  }
                                              }

            Returns:
                list - list of the appropriate JSON for an agent to send to the POST Subclient API
        """
        content = []

        try:
            for account in subclient_content:
                temp_content_dict = {
                    "cloudconnectorContent": {
                        "includeAccounts": {
                            "contentValue": account['display_name'],
                            "contentType": AppIDAType.CLOUD_APP.value,
                            "contentName": account['SMTPAddress']
                        }
                    }
                }

                content.append(temp_content_dict)
        except KeyError as err:
            raise SDKException('Subclient', '102', '{} not given in content'.format(err))

        self._set_subclient_properties("_content", content)

    def restore_out_of_place(
            self,
            client,
            destination_path,
            paths,
            overwrite=True,
            restore_data_and_acl=True,
            copy_precedence=None,
            from_time=None,
            to_time=None,
            to_disk=False):
        """Restores the files/folders specified in the input paths list to the input client,
            at the specified destionation location.

            Args:
                client                (str/object) --  either the name of the client or
                                                           the instance of the Client

                destination_path      (str)        --  full path of the restore location on client

                paths                 (list)       --  list of full paths of
                                                           files/folders to restore

                overwrite             (bool)       --  unconditional overwrite files during restore
                    default: True

                restore_data_and_acl  (bool)       --  restore data and ACL files
                    default: True

                copy_precedence         (int)   --  copy precedence value of storage policy copy
                    default: None

                from_time           (str)       --  time to retore the contents after
                        format: YYYY-MM-DD HH:MM:SS

                    default: None

                to_time           (str)         --  time to retore the contents before
                        format: YYYY-MM-DD HH:MM:SS

                    default: None

                to_disk             (bool)       --  If True, restore to disk will be performed

            Returns:
                object - instance of the Job class for this restore job

            Raises:
                SDKException:
                    if client is not a string or Client instance

                    if destination_path is not a string

                    if paths is not a list

                    if failed to initialize job

                    if response is empty

                    if response is not success

        """
        self._instance_object._restore_association = self._subClientEntity

        return self._instance_object.restore_out_of_place(
            client=client,
            destination_path=destination_path,
            paths=paths,
            overwrite=overwrite,
            restore_data_and_acl=restore_data_and_acl,
            copy_precedence=copy_precedence,
            from_time=from_time,
            to_time=to_time,
            to_disk=to_disk
        )

    def discover(self, discover_type='USERS'):
        """This method discovers the users/groups on OneDrive

                Args:

                    discover_type (str)  --  Type of discovery

                        Valid Values are

                        -   USERS
                        -   GROUPS

                        Default: USERS

                Returns:

                    List (list)  --  List of users on GSuite account

                Raises:
                    SDKException:
                        if response is empty

                        if response is not success


        """

        if discover_type.upper() == 'USERS':
            disc_type = 10
        elif discover_type.upper() == 'GROUPS':
            disc_type = 5
        _get_users = self._services['GET_CLOUDAPPS_USERS'] % (self._instance_object.instance_id,
                                                              self._client_object.client_id,
                                                              disc_type)

        flag, response = self._cvpysdk_object.make_request('GET', _get_users)
        if flag:
            if response.json() and "scDiscoveryContent" in response.json():
                self._discover_properties = response.json()[
                    "scDiscoveryContent"][0]

                if "contentInfo" in self._discover_properties:
                    self._contentInfo = self._discover_properties["contentInfo"]
                return self._contentInfo
            else:
                raise SDKException('Response', '102')
        else:
            raise SDKException('Response', '101', self._update_response_(response.text))

    def set_auto_discovery(self, value):
        """Sets the auto discovery value for subclient.
        You can either set a RegEx value or a user group,
        depending on the auto discovery type selected at instance level.

            Args:

                value   (list)  --  List of RegEx or user groups

        """

        if not isinstance(value, list):
            raise SDKException('Subclient', '116')

        if not self._instance_object.auto_discovery_status:
            raise SDKException('Subclient', '117')

        subclient_prop = self._subclient_properties['cloudAppsSubClientProp'].copy()
        if self._instance_object.auto_discovery_mode == 0:
            # RegEx based auto discovery is enabled on instance

            subclient_prop['oneDriveSubclient']['regularExp'] = value
            self._set_subclient_properties("_subclient_properties['cloudAppsSubClientProp']", subclient_prop)
        else:
            # User group based auto discovery is enabled on instance
            grp_list = []
            groups = self.discover(discover_type='GROUPS')
            for item in value:
                for group in groups:
                    if group['contentName'].lower() == item.lower():
                        grp_list.append({
                            "cloudconnectorContent": {
                                "includeAccounts": group
                            }
                        })
            self._content.extend(grp_list)
            self._set_subclient_properties("_subclient_properties['content']", self._content)
        self.refresh()

    def run_subclient_discovery(self):
        """
            This method launches AutoDiscovery on the subclient
        """

        discover_type = 15
        discover_users = self._services['GET_CLOUDAPPS_ONEDRIVE_USERS'] % (self._instance_object.instance_id,
                                                                           self._client_object.client_id,
                                                                           discover_type,
                                                                           self.subclient_id)
        flag, response = self._cvpysdk_object.make_request('GET', discover_users)
        if response.status_code != 200 and response.status_code != 500:
            raise SDKException('Response', '101')

    def add_AD_group(self, value):
        """Adds the user group to the subclient if auto discovery type selected
            AD group at instance level.
                Args:
                    value   (list)  --  List of user groups
        """
        grp_list = []
        groups = self.discover(discover_type='GROUPS')
        for item in value:
            for group in groups:
                if group['contentName'].lower() == item.lower():
                    grp_list.append(group)

        contentinfo = []

        for grp in grp_list:
            info = {
                "contentValue": grp['contentValue'],
                "contentType": grp['contentType'],
                "contentName": grp['contentName']
            }
            contentinfo.append(info)

        request_json = {
            "App_DiscoveryContent": {
                "scDiscoveryContent": [
                    {
                        "scEntity": {
                            "subclientId": self.subclient_id
                        },
                        "contentInfo": contentinfo
                    }
                ]
            }
        }
        add_ADgroup = self._services['EXECUTE_QCOMMAND']
        flag, response = self._cvpysdk_object.make_request('POST', add_ADgroup, request_json)

        if flag:
            if response.json() and 'errorCode' in response.json():
                error_code = response.json().get('errorCode')
                if error_code != 0:
                    raise SDKException('Response', '101')
            else:
                raise SDKException('Response', '102')
        else:
            raise SDKException('Response', '101', self._update_response_(response.text))

    def add_user(self, user_name):
        """This method adds one drive user to the subclient
                Args:
                    user_name   (str)  --  Onedrive user name
        """
        users = self.discover(discover_type='USERS')

        for user in users:
            if user['contentName'].lower() == user_name.lower():
                user_dict = user
                break

        request_json = {
            "App_DiscoveryContent": {
                "scDiscoveryContent": [
                    {
                        "scEntity": {
                            "subclientId": self.subclient_id
                        },
                        "contentInfo": [
                            {
                                "contentValue": user_dict['contentValue'],
                                "contentType": user_dict['contentType'],
                                "contentName": user_dict['contentName']
                            }
                        ]
                    }
                ]
            }
        }

        add_user = self._services['EXECUTE_QCOMMAND']
        flag, response = self._cvpysdk_object.make_request('POST', add_user, request_json)

        if flag:
            if response.json() and 'errorCode' in response.json():
                error_code = response.json().get('errorCode')
                if error_code != 0:
                    error_message = response.json().get('errorMessage')
                    output_string = 'Failed to user to the subclient\nError: "{0}"'
                    raise SDKException('Subclient', '102', output_string.format(error_message))
            else:
                raise SDKException('Response', '102')
        else:
            raise SDKException('Response', '101', self._update_response_(response.text))

    def _get_subclient_users(self):
        """Method to get the users in the subclient

            Returns:
                List of Users in subclient
        """
        users = []
        result = self.content
        for user in result:
            users.append(user['SMTPAddress'])
        return users

    @property
    def get_subclient_users(self):
        """Returns the users in subclient"""
        return self._get_subclient_users()

    def add_ad_group_onedrive_for_business_client(self,value,plan_name):
        """ Adds given OneDrive group to v2 client

            Args:

                value (string) : Group name

                plan_name (str) : O365 plan name to associate with users

            Raises:

                SDKException:

                    if response is not success

                    if response is returned with errors
        """
        # Get o365plan
        plan_name = plan_name.strip()
        o365_plan_object = self._commcell_object.plans.get(plan_name)
        o365_plan_id = int(o365_plan_object.plan_id)

        # Get client id
        client_id = int(self._client_object.client_id)

        groups = []
        group_response = self.search_for_group(group_id=value)
        display_name = group_response[0].get('name')
        group_id = group_response[0].get('id')

        groups.append({
            "name": display_name,
            "id": group_id
        })

        request_json = {
            "LaunchAutoDiscovery": True,
            "cloudAppAssociation": {
                "accountStatus": 0,
                "subclientEntity": {
                    "subclientId": int(self.subclient_id),
                    "clientId": client_id,
                    "applicationId": AppIDAType.CLOUD_APP.value
                },
                "cloudAppDiscoverinfo": {
                    "discoverByType": 2,
                    "groups": groups
                },
                "plan": {
                    "planId": o365_plan_id
                }
            }
        }

        user_associations = self._services['UPDATE_USER_POLICY_ASSOCIATION']
        flag, response = self._cvpysdk_object.make_request('POST', user_associations, request_json)

        if flag:
            if response.json() and 'errorCode' in response.json():
                error_code = response.json().get('errorCode')
                if error_code != 0:
                    error_message = response.json().get('errorMessage')
                    output_string = f'Failed to add group\nError: {error_message}'
                    raise SDKException('Subclient', '102', output_string)
        else:
            raise SDKException('Response', '101', self._update_response_(response.text))

    def add_users_onedrive_for_business_client(self, users, plan_name):
        """ Adds given OneDrive users to v2 client

            Args:

                users (list) : List of user's SMTP address

                plan_name (str) : O365 plan name to associate with users

            Raises:

                SDKException:

                    if response is not success

                    if response is returned with errors
        """

        if not (isinstance(users, list) and isinstance(plan_name, str)):
            raise SDKException('Subclient', '101')

        # Get o365plan
        plan_name = plan_name.strip()
        o365_plan_object = self._commcell_object.plans.get(plan_name)
        o365_plan_id = int(o365_plan_object.plan_id)

        # Get client ID
        client_id = int(self._client_object.client_id)

        user_accounts = []

        for user_id in users:
            # Get user details
            user_response = self.search_for_user(user_id)
            display_name = user_response[0].get('displayName')
            user_guid = user_response[0].get('user').get('userGUID')
            is_auto_discovered_user = user_response[0].get('isAutoDiscoveredUser')
            is_super_admin = user_response[0].get('isSuperAdmin')

            user_accounts.append({
                "displayName": display_name,
                "isSuperAdmin": is_super_admin,
                "smtpAddress": user_id,
                "isAutoDiscoveredUser": is_auto_discovered_user,
                "associated": False,
                "commonFlags": 0,
                "user": {
                    "userGUID": user_guid
                }
            })

        request_json = {
            "LaunchAutoDiscovery": False,
            "cloudAppAssociation": {
                "accountStatus": 0,
                "subclientEntity": {
                    "subclientId": int(self.subclient_id),
                    "clientId": client_id,
                    "applicationId": AppIDAType.CLOUD_APP.value
                },
                "cloudAppDiscoverinfo": {
                    "discoverByType": 1,
                    "userAccounts": user_accounts
                },
                "plan": {
                    "planId": o365_plan_id
                }
            }
        }

        user_associations = self._services['UPDATE_USER_POLICY_ASSOCIATION']
        flag, response = self._cvpysdk_object.make_request('POST', user_associations, request_json)

        if flag:
            if response.json() and 'errorCode' in response.json():
                error_code = response.json().get('errorCode')
                if error_code != 0:
                    error_message = response.json().get('errorMessage')
                    output_string = f'Failed to add user\nError: {error_message}'
                    raise SDKException('Subclient', '102', output_string)
        else:
            raise SDKException('Response', '101', self._update_response_(response.text))

    def verify_discovery_onedrive_for_business_client(self):
        """ Verifies that discovery is complete

            Returns:

                discovery_stats (tuple):

                    discovery_status (bool): True if users are discovered else returns False

                    total_records (int):     Number of users fetched, returns -1 if discovery is not complete

            Raises:

                 SDKException:

                        if response is not success

                        if response received does not contain pagining info
        """

        browse_content = (self._services['CLOUD_DISCOVERY'] % (self._instance_object.instance_id,
                                                               self._client_object.client_id,
                                                               AppIDAType.CLOUD_APP.value))

        # determines the number of accounts to return in response
        page_size = 1
        discover_query = f'{browse_content}&pageSize={page_size}'

        flag, response = self._cvpysdk_object.make_request('GET', discover_query)

        if flag:
            no_of_records = -1
            if response and response.json():
                if 'pagingInfo' in response.json():
                    no_of_records = response.json().get('pagingInfo', {}).get('totalRecords', -1)
                    if no_of_records > 0:
                        return True, no_of_records
            return False, no_of_records
        else:
            raise SDKException('Response', '101', self._update_response_(response.text))

    def search_for_user(self, user_id):
        """ Searches for a specific user's details from discovered list

            Args:
                user_id (str) : user's SMTP address

            Returns:

                user_accounts (list): user details' list fetched from discovered content
                              eg: [
                                      {
                                        'displayName': '',
                                        'smtpAddress': '',
                                        'isSuperAdmin': False,
                                        'isAutoDiscoveredUser': False,
                                        'commonFlags': 0,
                                        'user': {
                                            '_type_': 13,
                                             'userGUID': 'UserGuid'
                                             }
                                       }
                                  ]

            Raises:

                SDKException:

                    if discovery is not complete

                    if invalid SMTP address is passed

                    if response is empty

                    if response is not success
        """
        browse_content = (self._services['CLOUD_DISCOVERY'] % (self._instance_object.instance_id,
                                                               self._client_object.client_id,
                                                               AppIDAType.CLOUD_APP.value))


        search_query = f'{browse_content}&search={user_id}'

        flag, response = self._cvpysdk_object.make_request('GET', search_query)

        if flag:
            if response and response.json():
                if 'userAccounts' in response.json():
                    user_accounts = response.json().get('userAccounts', [])
                    if len(user_accounts) == 0:
                        error_string = 'Either discovery is not complete or user is not available in discovered data'
                        raise SDKException('Subclient', '102', error_string)
                    return user_accounts
                else:
                    raise SDKException('Response', '102', 'Check if the user provided is valid')
            else:
                raise SDKException('Response', '102')
        else:
            raise SDKException('Response', '101', self._update_response_(response.text))

    def search_for_group(self, group_id):
        """ Searches for a specific group details from discovered list

            Args:
                group_id (str) : group name

            Returns:

                groups (list): group details' list fetched from discovered content
                              eg: [
                                      {
                                         "name": "g1",
                                         "id": "df1f794a-ceee-4e91-b644-6f34a0416917"
                                       }
                                  ]

            Raises:

                SDKException:

                    if discovery is not complete

                    if invalid SMTP address is passed

                    if response is empty

                    if response is not success
        """
        browse_content = (self._services['GET_CLOUDAPPS_USERS'] % (self._instance_object.instance_id,
                                                              self._client_object.client_id,
                                                              5))

        search_query = f'{browse_content}&search={group_id}'

        flag, response = self._cvpysdk_object.make_request('GET', search_query)

        if flag:
            if response and response.json():
                if 'groups' in response.json():
                    groups = response.json().get('groups', [])
                    if len(groups) == 0:
                        error_string = 'Either discovery is not complete or group is not available in discovered data'
                        raise SDKException('Subclient', '102', error_string)
                    return groups
                else:
                    raise SDKException('Response', '102', 'Check if the group provided is valid')
            else:
                raise SDKException('Response', '102')
        else:
            raise SDKException('Response', '101', self._update_response_(response.text))

    def disk_restore_onedrive_for_business_client(self, users, destination_client, destination_path, skip_file_permissions=False):
        """ Runs an out-of-place restore job for specified users on OneDrive for business client
            By default restore skips the files already present in destination

            Args:
                users (list) : list of SMTP addresses of users
                destination_client (str) : client where the users need to be restored
                destination_path (str) : Destination folder location
                skip_file_permissions (bool) : If True, restore of file permissions are skipped (default: False)

            Returns:
                object - instance of the Job class for this restore job
        """
        self._instance_object._restore_association = self._subClientEntity
        source_user_list = self._get_user_guids(users)
        kwargs = {
            'disk_restore': True,
            'destination_path': destination_path,
            'destination_client': destination_client,
            'skip_file_permissions': skip_file_permissions
        }
        restore_json = self._instance_object._prepare_restore_json_onedrive_for_business_client(source_user_list, **kwargs)
        return self._process_restore_response(restore_json)

    def out_of_place_restore_onedrive_for_business_client(self, users, destination_path, **kwargs):
        """ Runs an out-of-place restore job for specified users on OneDrive for business client
            By default restore skips the files already present in destination

            Args:
                users (list) : list of SMTP addresses of users
                destination_path (str) : SMTP address of destination user
                **kwargs (dict) : Additional parameters
                    overwrite (bool) : unconditional overwrite files during restore (default: False)
                    restore_as_copy (bool) : restore files as copy during restore (default: False)
                    skip_file_permissions (bool) : If True, restore of file permissions are skipped (default: False)

            Returns:
                object - instance of the Job class for this restore job

            Raises:
                SDKException:

                    if overwrite and restore as copy file options are both selected
        """
        overwrite = kwargs.get('overwrite', False)
        restore_as_copy = kwargs.get('restore_as_copy', False)
        skip_file_permissions = kwargs.get('skip_file_permissions', True)
        include_deleted_items = kwargs.get('include_deleted_items',False)

        if overwrite and restore_as_copy:
            raise SDKException('Subclient', '102', 'Either select overwrite or restore as copy for file options')

        self._instance_object._restore_association = self._subClientEntity
        source_user_list = self._get_user_guids(users)
        kwargs = {
            'out_of_place': True,
            'destination_path': destination_path,
            'overwrite': overwrite,
            'restore_as_copy': restore_as_copy,
            'skip_file_permissions': skip_file_permissions,
            'include_deleted_items': include_deleted_items
        }
        restore_json = self._instance_object._prepare_restore_json_onedrive_for_business_client(source_user_list, **kwargs)

        return self._process_restore_response(restore_json)

    def in_place_restore_onedrive_syntex(self, users):
        """ Runs an in-place restore job for specified users on Syntex OneDrive for business client

            Args:
                users (list) :  List of SMTP addresses of users

            Returns:
                object - instance of the Job class for this restore job

            Raises:
                SDKException:

                    if restore job failed

                    if response is empty

                    if response is not success
        """

        user_details = {}
        for user in users:
            user_details[user] = self.search_for_user(user)

        self._instance_object._restore_association = self._subClientEntity

        syntex_restore_items = []
        for key, value in user_details.items():
            syntex_restore_items.append({
                "displayName": value[0]["displayName"],
                "email": value[0]["smtpAddress"],
                "guid": value[0]["user"]["userGUID"],
                "rawId": value[0]["user"]["userGUID"],
                "restoreType": 1
            })

        source_user_list = self._get_user_guids(users)
        restore_json = self._instance_object._prepare_restore_json_onedrive_for_business_client(source_user_list)

        # Get the current time in UTC
        current_time = datetime.datetime.now(datetime.timezone.utc)
        current_timestamp = int(current_time.timestamp())
        current_iso_format = current_time.strftime('%Y-%m-%dT%H:%M:%S.%f')[:-3] + 'Z'

        restore_json["taskInfo"]["subTasks"][0]["options"]["restoreOptions"]["cloudAppsRestoreOptions"][
            "msSyntexRestoreOptions"] = {
            "msSyntexRestoreItems": {
                "listMsSyntexRestoreItems": syntex_restore_items
            },
            "restoreDate": {
                "time": current_timestamp,
                "timeValue": current_iso_format
            },
            "restorePointId": "",
            "restoreType": 1,
            "useFastRestorePoint": True
        }

        return self._process_restore_response(restore_json)

    def in_place_restore_onedrive_for_business_client(self, users, **kwargs):
        """ Runs an in-place restore job for specified users on OneDrive for business client
            By default restore skips the files already present in destination

            Args:
                users (list) :  List of SMTP addresses of users
                **kwargs (dict) : Additional parameters
                    overwrite (bool) : unconditional overwrite files during restore (default: False)
                    restore_as_copy (bool) : restore files as copy during restore (default: False)
                    skip_file_permissions (bool) : If True, restore of file permissions are skipped (default: False)

            Returns:
                object - instance of the Job class for this restore job

            Raises:
                SDKException:

                    if overwrite and restore as copy file options are both selected
        """
        overwrite = kwargs.get('overwrite', False)
        restore_as_copy = kwargs.get('restore_as_copy', False)
        skip_file_permissions = kwargs.get('skip_file_permissions', True)
        include_deleted_items = kwargs.get('include_deleted_items',False)
        if overwrite and restore_as_copy:
            raise SDKException('Subclient', '102', 'Either select overwrite or restore as copy for file options')

        self._instance_object._restore_association = self._subClientEntity
        source_user_list = self._get_user_guids(users)
        kwargs = {
            'overwrite': overwrite,
            'restore_as_copy': restore_as_copy,
            'skip_file_permissions': skip_file_permissions,
            'include_deleted_items': include_deleted_items
        }

        restore_json = self._instance_object._prepare_restore_json_onedrive_for_business_client(source_user_list, **kwargs)
        return self._process_restore_response(restore_json)



    def _get_user_guids(self, users):
        """ Retrieve GUIDs for users specified

            Args:
                user (list) : List of SMTP addresses of users

            Returns:
                user_guid_list (list) : list of GUIDs of specified users

            Raises:
                SDKException:
                    if user details couldn't be found in discovered data
        """
        user_guid_list = []
        for user_id in users:
            user = self.search_for_user(user_id)
            if len(user) != 0 and user[0].get('user', {}).get('userGUID') is not None:
                user_guid_list.append(user[0].get('user').get('userGUID'))
            else:
                raise SDKException('Subclient', '102', 'User details not found in discovered data')
        return user_guid_list

    def process_index_retention_rules(self, index_app_type_id, index_server_client_name):
        """
         Makes API call to process index retention rules

         Args:

            index_app_type_id           (int)   --   index app type id

            index_server_client_name    (str)   --  client name of index server

         Raises:

                SDKException:

                    if index server not found

                    if response is empty

                    if response is not success
        """
        if self._commcell_object.clients.has_client(index_server_client_name):
            index_server_client_id = int(self._commcell_object.clients[index_server_client_name.lower()]['id'])
            request_json = {
                "appType": index_app_type_id,
                "indexServerClientId": index_server_client_id
            }
            flag, response = self._cvpysdk_object.make_request(
                'POST', self._services['OFFICE365_PROCESS_INDEX_RETENTION_RULES'], request_json
            )
            if flag:
                if response.json():
                    if "resp" in response.json():
                        error_code = response.json()['resp']['errorCode']
                        if error_code != 0:
                            error_string = response.json()['response']['errorString']
                            o_str = 'Failed to process index retention rules\nError: "{0}"'.format(error_string)
                            raise SDKException('Subclient', '102', o_str)
                    elif 'errorMessage' in response.json():
                        error_string = response.json()['errorMessage']
                        o_str = 'Failed to process index retention rules\nError: "{0}"'.format(error_string)
                        raise SDKException('Subclient', '102', o_str)
            else:
                raise SDKException('Response', '101', self._update_response_(response.text))
        else:
            raise SDKException('IndexServers', '102')

    def point_in_time_in_place_restore_onedrive_for_business_client(self, users, end_time, **kwargs):
        """ Runs an in-place point in time restore job for specified users on OneDrive for business client
            By default restore skips the files already present in destination

            Args:
                users (list) :  List of SMTP addresses of users
                end_time (int) : Backup job end time
                **kwargs (dict) : Additional parameters
                    overwrite (bool) : unconditional overwrite files during restore (default: False)
                    restore_as_copy (bool) : restore files as copy during restore (default: False)
                    skip_file_permissions (bool) : If True, restore of file permissions are skipped (default: False)
            Returns:
                object - instance of the Job class for this restore job

            Raises:
                SDKException:

                    if overwrite and restore as copy file options are both selected
        """

        overwrite = kwargs.get('overwrite', False)
        restore_as_copy = kwargs.get('restore_as_copy', False)
        skip_file_permissions = kwargs.get('skip_file_permissions', False)

        if overwrite and restore_as_copy:
            raise SDKException('Subclient', '102', 'Either select overwrite or restore as copy for file options')

        self._instance_object._restore_association = self._subClientEntity
        source_user_list = self._get_user_guids(users)
        kwargs = {
            'overwrite': overwrite,
            'restore_as_copy': restore_as_copy,
            'skip_file_permissions': skip_file_permissions
        }

        restore_json = self._instance_object._prepare_restore_json_onedrive_for_business_client(source_user_list, **kwargs)

        adv_search_bkp_time_dict={
                "field": "BACKUPTIME",
                "fieldValues": {
                    "values": [
                        "0",
                        str(end_time)
                    ]
                },
                "intraFieldOp": "FTOr"
            }


        add_to_time=restore_json["taskInfo"]["subTasks"][0]["options"]["restoreOptions"]["browseOption"]
        add_to_time["timeRange"]={"toTime":end_time}
        add_backup_time=restore_json["taskInfo"]["subTasks"][0]["options"]["restoreOptions"]["cloudAppsRestoreOptions"]["googleRestoreOptions"]["findQuery"]["advSearchGrp"]["fileFilter"][0]["filter"]["filters"]
        add_backup_time.append(adv_search_bkp_time_dict)

        return self._instance_object._process_restore_response(restore_json)

    def point_in_time_out_of_place_restore_onedrive_for_business_client(self, users, end_time, destination_path, **kwargs):
        """ Runs an out-of-place point in time restore job for specified users on OneDrive for business client
            By default restore skips the files already present in destination

            Args:
                users (list) : list of SMTP addresses of users
                end_time (int) : Backup job end time
                destination_path (str) : SMTP address of destination user
                **kwargs (dict) : Additional parameters
                    overwrite (bool) : unconditional overwrite files during restore (default: False)
                    restore_as_copy (bool) : restore files as copy during restore (default: False)
                    skip_file_permissions (bool) : If True, restore of file permissions are skipped (default: False)

            Returns:
                object - instance of the Job class for this restore job

            Raises:
                SDKException:

                    if overwrite and restore as copy file options are both selected
        """
        overwrite = kwargs.get('overwrite', False)
        restore_as_copy = kwargs.get('restore_as_copy', False)
        skip_file_permissions = kwargs.get('skip_file_permissions', False)

        if overwrite and restore_as_copy:
            raise SDKException('Subclient', '102', 'Either select overwrite or restore as copy for file options')

        self._instance_object._restore_association = self._subClientEntity
        source_user_list = self._get_user_guids(users)
        kwargs = {
            'out_of_place': True,
            'destination_path': destination_path,
            'overwrite': overwrite,
            'restore_as_copy': restore_as_copy,
            'skip_file_permissions': skip_file_permissions
        }
        restore_json = self._instance_object._prepare_restore_json_onedrive_for_business_client(source_user_list, **kwargs)

        adv_search_bkp_time_dict = {
            "field": "BACKUPTIME",
            "fieldValues": {
                "values": [
                    "0",
                    str(end_time)
                ]
            },
            "intraFieldOp": "FTOr"
        }

        add_to_time=restore_json["taskInfo"]["subTasks"][0]["options"]["restoreOptions"]["browseOption"]
        add_to_time["timeRange"]={"toTime":end_time}
        add_backup_time=restore_json["taskInfo"]["subTasks"][0]["options"]["restoreOptions"]["cloudAppsRestoreOptions"]["googleRestoreOptions"]["findQuery"]["advSearchGrp"]["fileFilter"][0]["filter"]["filters"]
        add_backup_time.append(adv_search_bkp_time_dict)

        return self._process_restore_response(restore_json)

    def run_user_level_backup_onedrive_for_business_client(self,users_list, custom_groups_list=[]):
        """
        Runs the backup for the users in users list/ custom categories list
        Args:
                users_list (list) : list of SMTP addresses of users
                custom_groups_list (lis) : list of custom categories

        Returns:
                object - instance of the Job class for this backup job
 
        Raises:
            SDKException:
                if response is empty
 
                if response is not success

        """
        task_json = self._task_json_for_onedrive_backup(users_list, custom_groups_list)
        create_task = self._services['CREATE_TASK']
        flag, response = self._commcell_object._cvpysdk_object.make_request(
            'POST', create_task, task_json
        )
        return self._process_backup_response(flag, response)

    def _get_user_details(self,user):
        """
        gets user details from discovery
        Args:
                user (str) : SMTP address of user

        Returns:
                user_details  (dict) : User's details fetched from discovery
 
        Raises:
            SDKException:
                if response is empty
        """
        user_details=self.search_for_user(user)
        if len(user_details)!=0:
            return user_details
        else:
            raise SDKException('Subclient', '102', 'User details not found in discovered data')

    def _get_group_details(self,group):
        """
        gets group details from discovery
        Args:
                group (str) : SMTP address of group
        """
        group_details=self.search_for_group(group)
        if len(group_details)!=0:
            return group_details
        else:
            raise SDKException('Subclient', '102', 'Group details not found in discovered data')

    def browse_for_content(self, discovery_type, include_deleted=False):
        """Returns the Onedrive client content i.e. users/ group information that is discovered in auto discovery phase

                Args:

                    discovery_type  (int)   --  type of discovery for content
                                                For all Associated users = 1
                                                For all Associated groups = 2
                                                For all Custom category groups = 31

                    include_deleted  (bool)  -- If True, deleted items will also be included

                Returns:

                    user_dict     (dict)    --  dictionary of users properties

                    no_of_records   (int)   --  no of records

                Raises:

                    SDKException:

                        if response is empty

                        if response is not success

                        if the method is called by Onedrive On-Premise Instance

        """
        if not self._backupset_object._instance_object.ca_instance_type.lower() == constants.ONEDRIVE_INSTANCE.lower():
            raise SDKException('Subclient', '102', 'Method not supported for Onedrive On-Premise Instance')

        self._USER_POLICY_ASSOCIATION = self._services['USER_POLICY_ASSOCIATION']
        request_json = {
            "discoverByType": discovery_type,
            "bIncludeDeleted": include_deleted,
            "cloudAppAssociation": {
                "subclientEntity": {
                    "subclientId": int(self.subclient_id)
                }
            },
            "searchInfo": {
                "isSearch": 0,
                "searchKey": ""
            }
        }
        flag, response = self._cvpysdk_object.make_request(
            'POST', self._USER_POLICY_ASSOCIATION, request_json
        )
        if flag:
            if response and response.json():
                no_of_records = 0
                if 'associations' in response.json():
                    no_of_records = response.json().get('associations', [{}])[0].get('pagingInfo', {}). \
                        get('totalRecords', -1)
                elif 'pagingInfo' in response.json():
                    no_of_records = response.json().get('pagingInfo', {}).get('totalRecords', -1)
                    if no_of_records <= 0:
                        return {}, no_of_records
                associations = response.json().get('associations', [{}])
                user_dict = {}
                if discovery_type == 2 or discovery_type == 31:
                    if associations:
                        for group in associations:
                            group_name = group.get("groups", {}).get("name", "")
                            user_dict[group_name] = {
                                'accountStatus': group.get("accountStatus"),
                                'discoverByType': group.get("discoverByType"),
                                'planName': group.get("plan", {}).get("planName", ""),
                                'id': group.get("groups", {}).get("id", ""),
                                'categoryNumber': group.get("groups", {}).get("categoryNumber", None)
                            }
                else:
                    if associations:
                        for user in associations:
                            user_url = user.get("userAccountInfo", {}).get("smtpAddress", "")
                            user_account_info = user.get("userAccountInfo", {})
                            user_dict[user_url] = {
                                'userAccountInfo': user_account_info,
                                'accountStatus': user.get("accountStatus"),
                                'discoverByType': user.get("discoverByType"),
                                'planName': user.get("plan", {}).get("planName", ""),
                                'lastBackupTime': user.get("userAccountInfo", {}).get("lastBackupJobRanTime", {}).get(
                                    "time", None)
                            }
                return user_dict, no_of_records
            return {}, 0
        raise SDKException('Response', '101', self._update_response_(response.text))

    def _set_properties_to_update_site_association(self, operation):
        """Updates the association properties of user

            Args:

               operation (int)                  --  type of operation to be performed
                                                     Example: 1 - Associate
                                                              2 - Enable
                                                              3 - Disable
                                                              4 - Remove

            Raises:

            SDKException:

                if the method is called by Onedrive On-Premise Instance

        """
        if not self._backupset_object._instance_object.ca_instance_type.lower() == constants.ONEDRIVE_INSTANCE.lower():
            raise SDKException('Subclient', '102', 'Method not supported for Onedrive On-Premise Instance')

        properties_dict = {}
        if operation == 1:
            properties_dict["accountStatus"] = 0
        elif operation == 2:
            properties_dict["accountStatus"] = 0
        elif operation == 3:
            properties_dict["accountStatus"] = 2
        elif operation == 4:
            properties_dict["accountStatus"] = 1
        return properties_dict


    def update_users_association_properties(self, operation, **kwargs):
        """Updates the association properties of user

                Args:
                    operation (int)         --  type of operation to be performed
                                                 Example: 1 - Associate
                                                          2 - Enable
                                                          3 - Disable
                                                          4 - Remove

                    Additional arguments (kwargs):
                    user_accounts_list (list)   --  list of user accounts
                                                    It has all information of users

                    groups_list (list)      --  list of groups
                                                It has all information of groups

                    plan_id (int)           --  id of Office 365 plan

                Raises:

                    SDKException:

                        if response is empty

                        if response is not success

                        if the method is called by Onedrive On-Premise Instance

        """
        plan_id = kwargs.get('plan_id', None)
        user_accounts_list = kwargs.get('user_accounts_list', None)
        groups_list = kwargs.get('groups_list', None)

        if not self._backupset_object._instance_object.ca_instance_type.lower() == constants.ONEDRIVE_INSTANCE.lower():
            raise SDKException('Subclient', '102', 'Method not supported for Onedrive On-Premise Instance')

        properties_dict = self._set_properties_to_update_site_association(operation)
        self._ASSOCIATE_CONTENT = self._services['UPDATE_USER_POLICY_ASSOCIATION']
        if user_accounts_list:
            request_json = {
                "cloudAppAssociation": {
                    "subclientEntity": {
                        "subclientId": int(self.subclient_id)
                    },
                    "cloudAppDiscoverinfo": {
                        "discoverByType": 1,
                        "userAccounts": user_accounts_list
                    }
                }
            }
        if groups_list:
            request_json = {
                "LaunchAutoDiscovery": True,
                "cloudAppAssociation": {
                    "subclientEntity": {
                        "subclientId": int(self.subclient_id)
                    },
                    "cloudAppDiscoverinfo": {
                        "discoverByType": 2,
                        "groups": groups_list
                    }
                }
            }
        if properties_dict.get('accountStatus', None) is not None:
            request_json['cloudAppAssociation']['accountStatus'] = properties_dict['accountStatus']
        if plan_id:
            request_json['cloudAppAssociation']['plan'] = {
                "planId": int(plan_id)
            }
        flag, response = self._cvpysdk_object.make_request(
            'POST', self._ASSOCIATE_CONTENT, request_json
        )
        if flag:
            if response.json():
                if "resp" in response.json():
                    error_code = response.json()['resp']['errorCode']
                    if error_code != 0:
                        error_string = response.json().get('response', {}).get('errorString', str())
                        o_str = 'Failed to associate content\nError: "{0}"'.format(error_string)
                        raise SDKException('Subclient', '102', o_str)
                elif 'errorMessage' in response.json():
                    error_string = response.json()['errorMessage']
                    o_str = 'Failed to associate content\nError: "{0}"'.format(error_string)
                    raise SDKException('Subclient', '102', o_str)
        else:
            raise SDKException('Response', '102', self._update_response_(response.text))

    def manage_custom_category(self, custom_dict, action, plan=None):
        """
        Adds or Edits Custom category in the office 365 app.

        Args:
            custom_dict (dict)  --  dictionary of custom category name and rule details.
                Example:
                    {
                    "name":"Display name contains custom"
                    "rules":
                        [
                            {
                            "CCRuleName":"User Display Name",
                            "CCRuleOperator":"Contains",
                            "CCRuleMask":"od_test_user"
                            }
                        ]
                    }
            action (str)     --  Action to perform. Either 'add' or 'edit'.
            plan (str)       --  Name of plan to be selected for adding category.
                                //Default: None. Required for adding category.

        Raises:
            SDKException:
                if response is not success
                if response is returned with errors
        """

        def get_field_number(field_name):
            """ Gets the indexed number for each type of field"""
            numbers = {
                "User Display Name": 1,
                "User SMTP Address": 2,
                "User Geo Location": 3,
                "License": 4
            }
            return numbers.get(field_name, None)

        def get_field_type(field_name):
            """ Returns the mapped field_type of given field_name """
            types = {
                "User Display Name": 5,
                "User SMTP Address": 5,
                "User Geo Location": 1,
                "License": 1
            }
            return types.get(field_name, None)

        def get_field_operator(cc_rule_operator):
            """ Gets the corresponding number assigned to each operator """
            operators = {
                "Contains": 0,
                "Regular Expression": 1,
                "Starts With": 3,
                "Ends With": 4,
                "Equals": 1000,
                "Not Equal": 1001
            }
            return operators.get(cc_rule_operator, None)

        def get_cc_rule_type(field_type):
            """  Gets the type of field in English words """
            if field_type == 1:
                return "Generic"
            elif field_type == 5:
                return "String"
            else:
                return "Unknown"

        def get_mask(cc_rule_mask, cc_rule_name):
            """ Gets the masked name of Custom category rule """
            if cc_rule_name != "User Geo Location":
                if cc_rule_name == "License":
                    if cc_rule_mask != "Active":
                        return "ActiveRevoked"
                return cc_rule_mask
            else:
                # Extract mask from the brackets in CCRuleMask
                match = re.search(r'\((.*?)\)', cc_rule_mask)
                if match:
                    return match.group(1)
                else:
                    return None

        # Get o365 plan object and ID
        if action == 'add':
            plan_name = plan.strip()
            o365_plan_object = self._commcell_object.plans.get(plan_name)
            o365_plan_id = int(o365_plan_object.plan_id)
        else:
            # Fetch plan details for the given category in case of edit
            groups, _ = self.browse_for_content(discovery_type=31)
            plan_name = groups[custom_dict['name']].get('planName', "")
            o365_plan_object = self._commcell_object.plans.get(plan_name)
            o365_plan_id = int(o365_plan_object.plan_id)
            categoryNumber = groups[custom_dict['name']].get('categoryNumber', None)

        # Get Instance, client, Subclient Ids
        instance_id = int(self._instance_object.instance_id)
        client_id = int(self._client_object.client_id)
        subclient_id = int(self.subclient_id)

        conditions = []
        for entry in custom_dict["rules"]:
            self.custom_counter += 1
            condition = {
                "uniqueId": f"CC_{self.custom_counter}",
                "fieldSource": "OD_KnownFields",
                "fieldName": entry["CCRuleName"],
                "fieldNumber": get_field_number(entry["CCRuleName"]),
                "fieldType": get_field_type(entry["CCRuleName"]),
                "fieldOperator": get_field_operator(entry["CCRuleOperator"]),
                "mask": get_mask(entry["CCRuleMask"], entry["CCRuleName"]),
                "CCRuleName": entry["CCRuleName"],
                "CCRuleOperator": entry["CCRuleOperator"],
                "CCRuleType": get_cc_rule_type(get_field_type(entry["CCRuleName"])),
                "CCRuleMask": entry["CCRuleMask"]
            }
            conditions.append(condition)

        req_json = {
            "subclientEntity": {
                "subclientId": subclient_id
            },
            "planEntity": {
                "planId": o365_plan_id,
                "planName": plan_name if action == 'edit' else ""
            },
            "status": 0,
            "categoryName": custom_dict['name'],
            "categoryQuery": {
                "conditions": conditions
            },
            "office365V2AutoDiscover": {
                "launchAutoDiscover": True,
                "appType": 134,
                "clientId": client_id,
                "instanceId": instance_id,
                "instanceType": 7
            }
        }

        if action == 'add':
            url = self._services['CUSTOM_CATEGORY'] % subclient_id
            flag, response = self._cvpysdk_object.make_request('POST', url, req_json)
        elif action == 'edit':
            url = self._services['CUSTOM_CATEGORY'] % subclient_id + "/" + str(categoryNumber)
            flag, response = self._cvpysdk_object.make_request('PUT', url, req_json)
        else:
            raise SDKException('Subclient', '102', "Invalid action. Must be either 'add' or 'edit'.")

        if flag:
            if response.json() and 'errorCode' in response.json():
                error_code = response.json().get('errorCode')
                if error_code != 0:
                    error_message = response.json().get('errorMessage')
                    output_string = f'Failed to {action} group\nError: {error_message}'
                    raise SDKException('Subclient', '102', output_string)
            else:
                raise SDKException('Response', '102')
        else:
            raise SDKException('Response', '101', self._update_response_(response.text))


    def update_custom_categories_association_properties(self, category_name, operation):
        """
        Updates the association properties of custom category

                Args:
                    category_name (str)     --  Display name of custom category
                    operation (int)         --  type of operation to be performed
                                                 Example:
                                                          0 - Enable
                                                          1 - Remove
                                                          2 - Disable

                Raises:

                    SDKException:

                        if response is empty

                        if response is not success

                        if the method is called by Onedrive On-Premise Instance

        """

        if not self._backupset_object._instance_object.ca_instance_type.lower() == constants.ONEDRIVE_INSTANCE.lower():
            raise SDKException('Subclient', '102', 'Method not supported for Onedrive On-Premise Instance')

        # Get Instance, client, Subclient Ids
        instance_id = int(self._instance_object.instance_id)
        client_id = int(self._client_object.client_id)
        client_name = self._client_object.client_name
        subclient_id = int(self.subclient_id)
        url = self._services['CUSTOM_CATEGORIES'] % subclient_id

        # Get the category number
        groups, numberOfGroups = self.browse_for_content(discovery_type=31)
        category_number = groups[category_name].get('categoryNumber', None)

        if not category_number:
            raise SDKException('Subclient', '102', 'Please ensure the category name given is valid')

        request_json = {
            "updateCategoryNumbers": [category_number],
            "subclientEntity": {
                "subclientId": subclient_id,
                "clientName": client_name
            },
            "office365V2AutoDiscover": {
                "launchAutoDiscover": True,
                "appType": 134,
                "clientId": client_id,
                "instanceId": instance_id
            },
            "status": operation
        }

        flag, response = self._cvpysdk_object.make_request(
            'PUT', url, request_json
        )

        if flag:
            if response.json() and 'errorCode' in response.json():
                error_code = response.json().get('errorCode')
                if error_code != 0:
                    error_message = response.json().get('errorMessage')
                    output_string = f'Failed to add group\nError: {error_message}'
                    raise SDKException('Subclient', '102', output_string)
            else:
                raise SDKException('Response', '102')
        else:
            raise SDKException('Response', '101', self._update_response_(response.text))

    def refresh_retention_stats(self, subclient_id):
        """
        refresh the retention stats for the client

        Args:
            subclient_id(int)             : subclient id of the client
        """
        request_json = {
            "appType": constants.ONEDRIVE_INDEX_APPTYPE_ID,
            "subclientId": int(subclient_id)
        }
        refresh_retention = self._services['OFFICE365_PROCESS_INDEX_RETENTION_RULES']
        flag, response = self._cvpysdk_object.make_request('POST', refresh_retention, request_json)

        if flag:
            if response.json() and 'errorCode' in response.json():
                error_code = response.json().get('errorCode')
                if error_code != 0:
                    error_message = response.json().get('errorMessage')
                    output_string = f'Failed to refresh retention stats \nError: {error_message}'
                    raise SDKException('Subclient', '102', output_string)
                else:
                    self.log.info("refresh retention stats successful")
        else:
            raise SDKException('Response', '101', self._update_response_(response.text))

    def refresh_client_level_stats(self,subclient_id):
        """
        refresh the client level stats for the client

        Args:
            subclient_id(int)             : subclient id of the client

        """
        request_json = {
            "appType": constants.ONEDRIVE_INDEX_APPTYPE_ID,
            "oneDriveIdxStatsReq":
                [{
                    "subclientId": int(subclient_id), "type": 0}]
        }
        refresh_backup_stats = self._services['OFFICE365_POPULATE_INDEX_STATS']
        flag, response = self._cvpysdk_object.make_request('POST', refresh_backup_stats, request_json)

        if flag:
            if response.json() and 'errorCode' in response.json():
                error_code = response.json().get('errorCode')
                if error_code != 0:
                    error_message = response.json().get('errorMessage')
                    output_string = f'Failed to refresh client level stats \nError: {error_message}'
                    raise SDKException('Subclient', '102', output_string)
                else:
                    self.log.info("refresh client level stats successful")
        else:
            raise SDKException('Response', '101', self._update_response_(response.text))

    def get_client_level_stats(self,backupset_id):
        """
        Returns the client level stats for the client

        Args:
            backupset_id(int)             : backupset id of the client

        Retruns:

            response(json)                : returns the client level stats as a json response
        """
        get_backup_stats = self._services['OFFICE365_OVERVIEW_STATS'] % backupset_id
        flag, response = self._cvpysdk_object.make_request('GET', get_backup_stats)

        if flag:
            if response.json() and 'errorCode' in response.json():
                error_code = response.json().get('errorCode')
                if error_code != 0:
                    error_message = response.json().get('errorMessage')
                    output_string = f'Failed to get client level stats \nError: {error_message}'
                    raise SDKException('Subclient', '102', output_string)
                else:
                    self.log.info("get client level stats successful")
        else:
            raise SDKException('Response', '101', self._update_response_(response.text))

        return response.json()







# -*- coding: utf-8 -*-

# --------------------------------------------------------------------------
# Copyright Commvault Systems, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# --------------------------------------------------------------------------

"""File for doing operations on an Storage Pools.

This module has classes defined for doing operations for Storage Pools:

StoragePools, StoragePoolType, StorageType, WORMLockType and StoragePool are the classes defined in this file.

StoragePools: Class for representing all the StoragePools in the commcell

StoragePoolType : Class for representing storage pool types like deduplication, secondary copy, non-dedupe, scale out

StorageType : Class for representing storage types like disk, cloud, tape of a storage pool

WORMLockType : Class for representing different WORM lock types flag values of a WORM enable storage pool

StoragePool: Class for representing a single StoragePool of the commcell


StoragePools
============

    __init__(commcell_object)   --  initializes object of the StoragePools class associated with the commcell

    __str__()                   --  returns all the storage pools associated with the commcell

    __repr__()                  --  returns the string representation of an instance of this class

    __len__()                   --  returns the number of storage pools added to the Commcell

    __getitem__()               --  returns the name of the storage pool for the given storage
    pool Id or the details for the given storage pool name

    _get_storage_pools()        --  returns all storage pools added to the commcell

    has_storage_pool()          --  checks whether the storage pool  with given name exists or not

    get()                       --  returns StoragePool object of the storage pool for the
                                    specified input name

    add()                       --  Adds a storage pool, according to given input and returns
                                    StoragePool object

    add_azure_storage_pool()    --  Adds new storage pool with provided name to the commcell

    delete()                    --  deletes the specified storage pool

    refresh()                   --  refresh the list of storage pools associated with the commcell

    add_air_gap_protect()       --  Adds a new air gap protect storage pool to commcell

Attributes
----------

    **all_storage_pools**   --  returns dict of all the storage pools on commcell


StoragePool
===========

    __init__()                  --  initialize the instance of StoragePool class for specific storage pool of commcell

    __repr__()                  --  returns a string representation of the StoragePool instance

     _get_storage_pool_properties() --  returns the properties of this storage pool

    refresh()                   --	Refresh the properties of the StoragePool

    get_copy()                  --  Returns the StoragePolicyCopy object of Storage Pool copy

    enable_compliance_lock()    --  Enables compliance lock on Storage Pool Copy

    enable_worm_storage_lock()  --  Enables WORM storage lock on storage pool

    hyperscale_add_nodes()      --  Add 3 new nodes to an existing storage pool

StoragePool instance attributes
================================

    **storage_pool_name**           --  returns the name of the storage pool

    **storage_pool_id**             --  returns the storage pool id

    **storage_pool_properties**     --  returns the properties of the storage pool

    **global_policy_name**          --  returns the global policy corresponding to the storage pool

    **copy_name**                   --  returns the copy name of the storage pool

    **copy_id**                     --  returns the copy id of the storage pool

    **storage_pool_type**           --  returns the storage pool type

    **storage_type**                --  returns the storage type of the storage pool

    **storage_vendor**              --  returns the storage vendor id of the storage pool

    **is_worm_storage_lock_enabled**--  returns whether WORM storage lock is enabled

    **is_object_level_worm_lock_enabled** --  returns whether object level WORM lock is enabled

    **is_bucket_level_worm_lock_enabled** --  returns whether bucket level WORM lock is enabled

    **is_compliance_lock_enabled**  --  returns whether compliance lock is enabled

# TODO: check with MM API team to get the response in JSON

"""
import copy

import xmltodict
from base64 import b64encode
from enum import IntFlag, IntEnum

from .exception import SDKException

from .storage import MediaAgent
from .security.security_association import SecurityAssociation
from .constants import StoragePoolConstants
from .policies.storage_policies import StoragePolicyCopy


class StoragePools:
    """Class for doing operations on Storage Pools, like get storage poo ID."""

    def __init__(self, commcell_object):
        """Initializes instance of the StoragePools class to perform operations on a storage pool.

            Args:
                commcell_object     (object)    --  instance of the Commcell class

            Returns:
                object  -   instance of the StoragePools class

        """
        self._commcell_object = commcell_object

        self._cvpysdk_object = commcell_object._cvpysdk_object
        self._services = commcell_object._services
        self._update_response_ = commcell_object._update_response_
        self._add_storage_pool_api = self._services['ADD_STORAGE_POOL']

        self._storage_pools_api = self._services['STORAGE_POOL']

        self._metallic_storage_api = self._services['GET_METALLIC_STORAGE_DETAILS']
        self._storage_pools = None

        self.refresh()

    def __str__(self):
        """Representation string consisting of all storage pools present in the Commcell.

            Returns:
                str     -   string of all the storage pools associated with the commcell

        """
        representation_string = '{:^5}\t{:^40}\n\n'.format('S. No.', 'Storage Pool')

        for index, storage_pool in enumerate(self._storage_pools):
            sub_str = '{:^5}\t{:40}\n'.format(index + 1, storage_pool)
            representation_string += sub_str

        return representation_string.strip()

    def __repr__(self):
        """Returns the string representation of an instance of this class."""
        return "StoragePools class instance for Commcell: '{0}'".format(
            self._commcell_object.commserv_name
        )

    def __len__(self):
        """Returns the number of the storage pools added to the Commcell."""
        return len(self.all_storage_pools)

    def __getitem__(self, value):
        """Returns the name of the storage pool for the given storage pool ID or
            the details of the storage pool for given storage pool Name.

            Args:
                value   (str / int)     --  Name or ID of the storage pool

            Returns:
                str     -   name of the storage pool, if the storage pool id was given

                dict    -   dict of details of the storage pool, if storage pool name was given

            Raises:
                IndexError:
                    no storage pool exists with the given Name / Id

        """
        value = str(value)

        if value in self.all_storage_pools:
            return self.all_storage_pools[value]
        else:
            try:
                return list(filter(lambda x: x[1]['id'] == value, self.all_storage_pools.items()))[0][0]
            except IndexError:
                raise IndexError('No storage pool exists with the given Name / Id')

    def _get_storage_pools(self):
        """Gets all the storage pools associated with the Commcell environment.

            Returns:
                dict    -   consists of all storage pools added to the commcell

                    {
                        "storage_pool1_name": storage_pool1_id,

                        "storage_pool2_name": storage_pool2_id
                    }

            Raises:
                SDKException:
                    if response is empty

                    if response is not success

        """
        headers = self._commcell_object._headers.copy()
        headers['Accept'] = 'application/xml'

        flag, response = self._cvpysdk_object.make_request(
            'GET', self._storage_pools_api, headers=headers
        )

        if flag:
            storage_pools = {}

            response = xmltodict.parse(response.text)['Api_GetStoragePoolListResp']

            if response is None or response.get('storagePoolList') is None:
                storage_pool_list = []
            else:
                storage_pool_list = response['storagePoolList']

            if not isinstance(storage_pool_list, list):
                storage_pool_list = [storage_pool_list]

            if response:
                for pool in storage_pool_list:
                    name = pool['storagePoolEntity']['@storagePoolName'].lower()
                    storage_pool_id = pool['storagePoolEntity']['@storagePoolId']

                    storage_pools[name] = storage_pool_id

                return storage_pools
        else:
            response_string = self._update_response_(response.text)
            raise SDKException('Response', '101', response_string)

    @property
    def all_storage_pools(self):
        """Returns dict of all the storage pools on this commcell

            dict    -   consists of all storage pools added to the commcell

                {

                    "storage_pool1_name": storage_pool1_id,

                    "storage_pool2_name": storage_pool2_id
                }

        """
        return self._storage_pools

    def has_storage_pool(self, name):
        """Checks if a storage pool exists in the Commcell with the input storage pool name.

            Args:
                name    (str)   --  name of the storage pool

            Returns:
                bool    -   boolean output whether the storage pool exists in the commcell or not

        """
        return self._storage_pools and name.lower() in self._storage_pools

    def get(self, name):
        """Returns the id of the storage pool for the given storage pool name.

            Args:
                name    (str)   --  name of the storage pool to get the id of

            Returns:
                str     -   id of the storage pool for the given storage pool name

            Raises:
                SDKException:
                    if no storage pool exists with the given name

        """
        name = name.lower()

        if self.has_storage_pool(name):
            return StoragePool(self._commcell_object, name, storage_pool_id=self._storage_pools[name])
        else:
            raise SDKException('StoragePool', '103')

    def hyperscale_create_storage_pool(self, storage_pool_name, media_agents):
        """
            Create new storage pool for hyperscale
            Args:
                storage_pool_name (string) -- Name of the storage pools to create

                media_agents      (List)   -- List of 3 media agents with name's(str)
                                                or instance of media agent's(object)

                Example: ["ma1","ma2","ma3"]

            Return:
                 flag, response -- response returned by the REST API call
        """

        if not isinstance(media_agents, list):
            raise SDKException('Storage', '101')
        if not isinstance(storage_pool_name, str):
            raise SDKException('Storage', '101')

        mediagent_obj = []
        for media_agent in media_agents:
            if isinstance(media_agent, MediaAgent):
                mediagent_obj.append(media_agent)
            elif isinstance(media_agent, str):
                mediagent_obj.append(self._commcell_object.media_agents.get(media_agent))
            else:
                raise SDKException('Storage', '103')
        if len(mediagent_obj) <= 2:
            raise SDKException('Storage', '102', "minimum 3 media agents are required")

        request_xml = """<App_CreateStoragePolicyReq storagePolicyName="{0}" copyName="{0}_Primary" type="1"
                                     numberOfCopies="1">
                                    <storagePolicyCopyInfo>
                                        <storagePolicyFlags scaleOutStoragePolicy="1"/>
                                    </storagePolicyCopyInfo>
                                    <storage>
                                        <mediaAgent mediaAgentId="{4}" mediaAgentName="{1}" displayName="{1}"/>
                                    </storage>
                                    <storage>
                                        <mediaAgent mediaAgentId="{5}" mediaAgentName="{2}" displayName="{2}"/>
                                    </storage>
                                    <storage>
                                        <mediaAgent mediaAgentId="{6}" mediaAgentName="{3}" displayName="{3}"/>
                                    </storage>
                                    <scaleoutConfiguration configurationType="1"/>
                                </App_CreateStoragePolicyReq>
                                """.format(storage_pool_name, mediagent_obj[0].media_agent_name,
                                           mediagent_obj[1].media_agent_name, mediagent_obj[2].media_agent_name,
                                           mediagent_obj[0].media_agent_id, mediagent_obj[1].media_agent_id,
                                           mediagent_obj[2].media_agent_id)

        flag, response = self._commcell_object._cvpysdk_object.make_request(
            'POST', self._add_storage_pool_api, request_xml
        )
        if flag:
            if response.json():
                error_code = response.json()['error']['errorCode']

                if int(error_code) != 0:
                    error_message = response.json()['error']['errorMessage']
                    o_str = 'Failed to create storage pool\nError: "{0}"'

                    raise SDKException('StoragePool', '102', o_str.format(error_message))
            else:
                raise SDKException('Response', '102')
        else:
            response_string = self._commcell_object._update_response_(response.text)
            raise SDKException('Response', '101', response_string)

        self.refresh()
        return self.get(storage_pool_name)

    def add_air_gap_protect(self, storage_pool_name, media_agent, storage_type, storage_class, region_name,
                            ddb_ma=None, dedup_path=None):
        """
            Adds a new air gap protect storage pool to commcell

                Args:
                    storage_pool_name   (str)       --  name of new storage pool to add

                    media_agent         (str/object)--  name or instance of media agent

                    storage_type        (str)        -- name of the cloud vendor (str, eg - "Microsoft Azure storage") (same as UI)

                    storage_class       (str)        -- storage class (str, eg - "Hot","Cool") (same as UI)

                    region_name (str)      --  name of the geographical region for storage (same as UI)

                    ddb_ma              (list<str/object>/str/object)   --  list of (name of name or instance)
                                                                            or name or instance of dedupe media agent

                    dedup_path          (list<str>/str)       --  list of paths or path where the DDB should be stored

                Returns:
                    StoragePool object if creation is successful

                Raises:
                    SDKException, if invalid parameters provided

        """
        license_type_dict = StoragePoolConstants.AIR_GAP_PROTECT_STORAGE_TYPES
        error_message = ""
        if storage_type.upper() in license_type_dict:
            available_storage_classes = license_type_dict[storage_type.upper()]
            if storage_class.upper() in available_storage_classes:
                vendor_id = available_storage_classes[storage_class.upper()]["vendorId"]
                display_vendor_id = available_storage_classes[storage_class.upper()]["displayVendorId"]
            else:
                error_message += f"Invalid storage class provided. Valid storage class {list(available_storage_classes.keys())}"
        else:
            error_message += f"  Invalid storage type provided. {list(license_type_dict.keys())}"

        if error_message:
            raise SDKException('Storage', '101', error_message)

        region = None
        available_regions = []

        #  API call to fetch the region name - sourced directly from the vendor
        flag, response = self._commcell_object._cvpysdk_object.make_request('GET', self._metallic_storage_api)
        if flag:
            if response.json():
                if "storageInformation" in response.json():
                    for storage_info in response.json()["storageInformation"]:
                        if (int(storage_info["vendorId"]) == int(vendor_id)) and (int(storage_info["displayVendorId"]) == int(display_vendor_id)):
                            for region_dict in storage_info["region"]:
                                available_regions.append(region_dict["displayName"])
                                if region_dict["displayName"] == region_name:
                                    region = region_dict["regionName"]
                                    break

                        if region:
                            break

                    if region is None:
                        if not available_regions:
                            raise SDKException('Storage', '101',
                                               f"Active license is required to configure {storage_type} - {storage_class} Air Gap Protect storage")
                        else:
                            raise SDKException('Storage', '101',
                                               f"Invalid region: {region_name} ,\nValid regions: {available_regions}")
                else:
                    raise SDKException('Storage', '101', "Unexpected response returned while fetching Air Gap Protect storage details")
            else:
                raise SDKException('Response', '102')
        else:
            response_string = self._commcell_object._update_response_(response.text)
            raise SDKException('Response', '101', response_string)

        # cloud server type for air gap protect is 400
        cloud_server_type = 400

        return self.add(storage_pool_name=storage_pool_name, mountpath=None, media_agent=media_agent, ddb_ma=ddb_ma,
                        dedup_path=dedup_path, cloud_server_type=cloud_server_type, region=region, vendor_id=vendor_id,
                        display_vendor_id=display_vendor_id)

    def add(self, storage_pool_name, mountpath, media_agent, ddb_ma=None, dedup_path=None, **kwargs):
        """
        Adds a new storage pool to commcell

        Args:
            storage_pool_name   (str)       --  name of new storage pool to add

            mountpath           (str)       --  mount path for the storage pool

            media_agent         (str/object)--  name or instance of media agent

            ddb_ma              (list<str/object>/str/object)   --  list of (name of name or instance)
                                                                        or name or instance of dedupe media agent

            dedup_path          (list<str>/str)       --  list of paths or path where the DDB should be stored

            **kwargs:
                username        (str)       --  username to access the mountpath

                password        (str)       --  password to access the mountpath

                credential_name (str)       --  name of the credential as in credential manager

                cloud_server_type (int)     --  cloud server type of the cloud vendor (required)

                region (str)                --  name of geographical region for storage (required for air gap protect)

                vendor_id (int)             -- id for the cloud_vendor (eg - 3 for azure) (required for air gap protect pool)

                display_vendor_id (int)     -- storage Class id for that vendor (eg - 401 for azure hot) (required for air gap protect pool)

                region_id        (int)      --  Cloud Hypervisor specific region ID

        Returns:
            StoragePool object if creation is successful

        Raises:
            Exception if creation is unsuccessful
        """
        username = kwargs.get('username', None)
        password = kwargs.get('password', None)
        credential_name = kwargs.get('credential_name', None)
        cloud_server_type = kwargs.get('cloud_server_type', None)
        library_name = kwargs.get('library_name', None)

        region = kwargs.get('region', None)
        vendor_id = kwargs.get('vendor_id', None)
        display_vendor_id = kwargs.get('display_vendor_id', None)
        region_id = kwargs.get('region_id', None)

        if ((ddb_ma is not None and not (isinstance(dedup_path, str) or isinstance(dedup_path, list))) or
                not (isinstance(storage_pool_name, str) or not isinstance(mountpath, str))):
            raise SDKException('Storage', '101')

        if isinstance(media_agent, MediaAgent):
            media_agent = media_agent
        elif isinstance(media_agent, str):
            media_agent = MediaAgent(self._commcell_object, media_agent)
        else:
            raise SDKException('Storage', '103')

        if (isinstance(ddb_ma, str) or isinstance(ddb_ma, MediaAgent)) and isinstance(dedup_path, str):
            ddb_ma = [ddb_ma]
            dedup_path = [dedup_path]

        if isinstance(ddb_ma, list) and isinstance(dedup_path, list):
            if len(ddb_ma) != len(dedup_path):
                raise SDKException('Storage', '101')

        if library_name is not None and mountpath != '':
            raise SDKException('Storage', '101')

        if ddb_ma is not None and (len(ddb_ma) > 6 or len(dedup_path) > 6):
            raise SDKException('Storage', '110')

        if ddb_ma is not None:
            for i in range(len(ddb_ma)):
                if isinstance(ddb_ma[i], MediaAgent):
                    ddb_ma[i] = ddb_ma[i]
                elif isinstance(ddb_ma[i], str):
                    ddb_ma[i] = MediaAgent(self._commcell_object, ddb_ma[i])
                else:
                    raise SDKException('Storage', '103')

        request_json = {
            "storagePolicyName": storage_pool_name,
            "type": "CVA_REGULAR_SP",
            "copyName": "Primary",
            "numberOfCopies": 1,
            "storage": [
                {
                    "path": mountpath,
                    "mediaAgent": {
                        "mediaAgentId": int(media_agent.media_agent_id),
                        "mediaAgentName": media_agent.media_agent_name
                    }
                }
            ],
            "storagePolicyCopyInfo": {
                "copyType": "SYNCHRONOUS",
                "isFromGui": True,
                "active": "SET_TRUE",
                "isDefault": "SET_TRUE",
                "numberOfStreamsToCombine": 1,
                "retentionRules": {
                    "retentionFlags": {
                        "enableDataAging": "SET_TRUE"
                    },
                    "retainBackupDataForDays": -1,
                    "retainBackupDataForCycles": -1,
                    "retainArchiverDataForDays": -1
                },
                "library": {
                    "libraryId": 0,
                },
                "mediaAgent": {
                    "mediaAgentId": int(media_agent.media_agent_id),
                    "mediaAgentName": media_agent.media_agent_name
                }
            }
        }

        if cloud_server_type and int(cloud_server_type) > 0:
            request_json["storage"][0]["deviceType"] = cloud_server_type

        if region_id is not None:
            request_json["storage"][0]["metallicStorageInfo"] = {
                "region": [
                    {
                        "regionId": region_id
                    }
                ],
                "storageClass": [
                    "CONTAINER_DEFAULT"
                ],
                "replication": [
                    "NONE"
                ]
            }
            request_json["region"] = {"regionId": region_id}

        if username is not None:
            request_json["storage"][0]["credentials"] = {"userName": username}

        if password is not None:
            request_json["storage"][0]["credentials"]["password"] = b64encode(password.encode()).decode()

        if credential_name is not None:
            request_json["storage"][0]["savedCredential"] = {"credentialName": credential_name}

        if library_name is not None:
            request_json["storage"] = []
            request_json["storagePolicyCopyInfo"]["library"]["libraryName"] = library_name

        if ddb_ma is not None or dedup_path is not None:
            maInfoList = []
            for ma, path in zip(ddb_ma, dedup_path):
                maInfoList.append({
                    "mediaAgent": {
                        "mediaAgentId": int(ma.media_agent_id),
                        "mediaAgentName": ma.media_agent_name
                    },
                    "subStoreList": [
                        {
                            "accessPath": {
                                "path": path
                            },
                            "diskFreeThresholdMB": 5120,
                            "diskFreeWarningThreshholdMB": 10240
                        }]
                })

            request_json["storagePolicyCopyInfo"].update({
                "storagePolicyFlags": {
                    "blockLevelDedup": "SET_TRUE",
                    "enableGlobalDeduplication": "SET_TRUE"
                },
                "dedupeFlags": {
                    "enableDeduplication": "SET_TRUE",
                    "enableDASHFull": "SET_TRUE",
                    "hostGlobalDedupStore": "SET_TRUE"
                },
                "DDBPartitionInfo": {
                    "maInfoList": maInfoList
                }
            })
        else:
            request_json["storagePolicyCopyInfo"].update({
                "storagePolicyFlags": {
                    "globalStoragePolicy": "SET_TRUE"
                },
                "copyFlags": {
                    "preserveEncryptionModeAsInSource": "SET_TRUE"
                },
                "extendedFlags": {
                    "globalStoragePolicy": "SET_TRUE"
                }
            })

        # air gap protect storage
        if cloud_server_type == 400:
            del request_json["storage"][0]["path"]
            request_json["storage"][0]["savedCredential"] = {"credentialId": 0}

            metallic_Storage = {
                "region": [
                    {
                        "regionName": region
                    }
                ],
                "displayVendorId": display_vendor_id,
                "vendorId": vendor_id
            }
            request_json["storage"][0]["metallicStorageInfo"] = metallic_Storage

        flag, response = self._commcell_object._cvpysdk_object.make_request(
            'POST', self._add_storage_pool_api, request_json
        )

        if flag:
            if response.json():
                error_code = response.json()['error']['errorCode']

                if int(error_code) != 0:
                    error_message = response.json()['error']['errorMessage']
                    o_str = 'Failed to create storage pool\nError: "{0}"'

                    raise SDKException('StoragePool', '102', o_str.format(error_message))
            else:
                raise SDKException('Response', '102')
        else:
            response_string = self._commcell_object._update_response_(response.text)
            raise SDKException('Response', '101', response_string)

        self.refresh()
        self._commcell_object.disk_libraries.refresh()
        return self.get(storage_pool_name)

    def add_azure_storage_pool(self, storage_pool_name, container_name, media_agents, dedup_paths, **kwargs):
        """ Adds new storage pool with provided name to the commcell
                 Args:
                     storage_pool_name (str)     --  name of the storage pool to be created

                     container_name (str)        --  container name to be used with storage pool

                     media_agents (list)         --  list of media agent names to be used for storage pool

                     dedup_paths (list)          --  list of paths for storing deduplication data

                     **kwargs (dict)             --  dict of keyword arguments as follows
                         username        (str)   --  azure storage credential username
                         password        (str)   --  azure storage credential password
                         credential_name (str)   --  Credential name to be used

                 Returns:
                     Azure storage policy object

                 Raises:
                     SDKException:
                         If invalid type arguments are passed
                         If Storage Pool with given name already exist
                         Response was not success.
                         Response was empty.

        """
        username = kwargs.get("username", "")
        password = kwargs.get("password", "")
        credential_name = kwargs.get("credential_name", "")
        if not (isinstance(storage_pool_name, str) and isinstance(container_name, str)
                and isinstance(media_agents, list) and isinstance(dedup_paths, list)
                and isinstance(username, str) and isinstance(password, str)
                and isinstance(credential_name, str)):
            raise SDKException('StoragePool', '101')
        storage_pools = self._commcell_object.storage_pools
        if storage_pools.has_storage_pool(storage_pool_name):
            raise SDKException('StoragePool', '103')
        request_json = copy.deepcopy(StoragePoolConstants.AZURE_STORAGE_REQ_JSON)
        request_json["storagePolicyName"] = storage_pool_name
        storage_policy_info = request_json["storagePolicyCopyInfo"]
        storage_ma = self._commcell_object.media_agents.get(media_agents[0])
        storage_policy_info["library"]["libraryName"] = container_name
        storage_policy_info["mediaAgent"]["mediaAgentId"] = int(storage_ma.media_agent_id)
        storage_policy_info["mediaAgent"]["mediaAgentName"] = storage_ma.media_agent_name
        ddb_info = []
        for (ma, ddb_path) in zip(media_agents, dedup_paths):
            ma_ddb = copy.deepcopy(StoragePoolConstants.MA_INFO_LIST)
            ma_info = self._commcell_object.media_agents.get(ma)
            ma_ddb["mediaAgent"]["mediaAgentId"] = int(ma_info.media_agent_id)
            ma_ddb["mediaAgent"]["mediaAgentName"] = ma_info.media_agent_name
            ma_ddb["subStoreList"][0]["accessPath"]["path"] = ddb_path
            ddb_info.append(ma_ddb)
        storage_policy_info["DDBPartitionInfo"]["maInfoList"] = ddb_info
        storage_info = request_json["storage"][0]
        storage_info["path"] = container_name
        storage_info["mediaAgent"]["mediaAgentId"] = int(storage_ma.media_agent_id)
        storage_info["mediaAgent"]["mediaAgentName"] = storage_ma.media_agent_name
        storage_info["credentials"]["userName"] = username
        storage_info["credentials"]["password"] = password
        credential = self._commcell_object.credentials.get(credential_name)
        storage_info["savedCredential"]["credentialId"] = credential.credential_id
        storage_info["savedCredential"]["credentialName"] = credential.credential_name
        flag, response = self._commcell_object._cvpysdk_object.make_request(
            'POST', self._add_storage_pool_api, request_json
        )

        if flag:
            if response.json():
                error_code = response.json().get('error', {}).get('errorCode', 0)
                if int(error_code) != 0:
                    error_message = response.json()['error']['errorMessage']
                    o_str = 'Failed to create storage policy\nError: "{0}"'

                    raise SDKException('StoragePool', '102', o_str.format(error_message))
            else:
                raise SDKException('Response', '102')
        else:
            response_string = self._commcell_object._update_response_(response.text)
            raise SDKException('Response', '101', response_string)

        self.refresh()
        return self.get(request_json["storagePolicyName"])


    def delete(self, storage_pool_name):
        """deletes the specified storage pool.

            Args:
                storage_pool_name (str)  --  name of the storage pool to delete

            Raises:
                SDKException:
                    if type of the storage pool name is not string

                    if failed to delete storage pool

                    if no storage pool exists with the given name

                    if response is empty

                    if response is not success

        """

        if not isinstance(storage_pool_name, str):
            raise SDKException('Storage', '101')
        else:
            storage_pool_name = storage_pool_name.lower()

            if self.has_storage_pool(storage_pool_name):
                storage_pool_id = self._storage_pools[storage_pool_name]

                delete_storage_pool = self._services['DELETE_STORAGE_POOL'] % (storage_pool_id)

                flag, response = self._cvpysdk_object.make_request('DELETE', delete_storage_pool)

                if flag:
                    error_code = response.json()['error']['errorCode']
                    if int(error_code) != 0:
                        error_message = response.json()['error']['errorMessage']
                        o_str = f'Failed to delete storage pools {storage_pool_name}'
                        o_str += '\nError: "{0}"'.format(error_message)
                        raise SDKException('Storage', '102', o_str)
                    else:
                        # initialize the storage pool again
                        # so the storage pool object has all the storage pools
                        self.refresh()
                        # as part of storage pool we might delete library so initialize the libraries again
                        self._commcell_object.disk_libraries.refresh()
                else:
                    response_string = self._update_response_(response.text)
                    raise SDKException('Response', '101', response_string)
            else:
                raise SDKException(
                    'Storage',
                    '102',
                    'No storage pool exists with name: {0}'.format(storage_pool_name)
                )
    def refresh(self):
        """Refresh the list of storage pools associated to the Commcell."""
        self._storage_pools = self._get_storage_pools()

class StoragePoolType(IntEnum):
    """Class IntEnum to represent different storage pool types"""
    DEDUPLICATION = 1,
    SECONDARY_COPY = 2,
    NON_DEDUPLICATION = 3,
    SCALE_OUT = 4


class StorageType(IntEnum):
    """Class IntEnum to represent different storage types"""
    DISK = 1,
    CLOUD = 2,
    HYPERSCALE = 3,
    TAPE = 4


class WORMLockType(IntFlag):
    """Class IntFlag to represent different WORM lock types flag values"""
    COPY = 1,  # copy level software WORM (compliance lock)
    STORAGE = 2,  # storage level hardware WORM
    OBJECT = 4,  # object level storage WORM
    BUCKET = 8  # bucket level storage WORM


class StoragePool(object):
    """Class for individual storage pools"""

    def __init__(self, commcell_object, storage_pool_name, storage_pool_id=None):
        """
        Intitalise the Storage Pool classs instance

        Args:
            commcell_object     (object)        --instance of the Commcell class

            storage_pool_name   (string)    -- Name of the storage pool

            storage_pool_id     (int)       -- Storage pool id
        Returns:
            object - Instance of the StoragePool class

        """
        self._storage_pool_name = storage_pool_name.lower()
        self._commcell_object = commcell_object
        self._storage_pool_properties = None
        self._storage_pool_id = None
        self._copy_id = None
        self._copy_name = None

        if storage_pool_id:
            self._storage_pool_id = str(storage_pool_id)
        else:
            self._storage_pool_id = self._commcell_object.storage_pools.get(self._storage_pool_name).storage_pool_id

        self._STORAGE_POOL = self._commcell_object._services['GET_STORAGE_POOL'] % (self.storage_pool_id)
        self.refresh()

        self._copy_id = self._storage_pool_properties.get("storagePoolDetails", {}).get("copyInfo", {}).get(
            "StoragePolicyCopy", {}).get("copyId")
        self._copy_name = self._storage_pool_properties.get("storagePoolDetails", {}).get("copyInfo", {}).get(
            "StoragePolicyCopy", {}).get("copyName")

    def __repr__(self):
        """String representation of the instance of this class"""
        representation_string = "Storage Pool class Instance for {0}".format(self._storage_pool_name)
        return representation_string

    def _get_storage_pool_properties(self):
        """
        Gets StoragePool properties

            Raises:
                SDKException:
                    if repsonse is empty

                    if response is not success

        """
        flag, response = self._commcell_object._cvpysdk_object.make_request('GET', self._STORAGE_POOL)

        if flag:
            if response.json():
                self._storage_pool_properties = response.json()
            else:
                raise SDKException('Response', '102')
        else:
            response_string = self._commcell_object._update_response_(response.text)
            raise SDKException('Response', '101', response_string)

    @property
    def storage_pool_name(self):
        """Treats the storage_policy_name as a read only attribute"""
        return self._storage_pool_name

    @property
    def storage_pool_id(self):
        """Treats id as a read only attribute"""
        return self._storage_pool_id

    @property
    def storage_pool_properties(self):
        """Treats the storage_pool_properties as a read only attribute"""
        return self._storage_pool_properties

    @property
    def global_policy_name(self):
        """Returns the global policy corresponding to the storage pool"""
        return self._storage_pool_properties["storagePoolDetails"]["copyInfo"]["StoragePolicyCopy"]["storagePolicyName"]

    @property
    def copy_name(self):
        """Treats copy name as a read only attribute"""
        return self._copy_name

    @property
    def copy_id(self):
        """Treats copy ID as a read only attribute"""
        return self._copy_id

    @property
    def storage_pool_type(self):
        """Treats storage type as a read only attribute"""
        return self._storage_pool_properties["storagePoolDetails"]["storagePoolType"]

    @property
    def storage_type(self):
        """Treats storage type as a read only attribute"""
        return self._storage_pool_properties["storagePoolDetails"]["storageType"]

    @property
    def storage_vendor(self):
        """Treats library vendor like cloud storage provider as a read only attribute"""
        return self._storage_pool_properties["storagePoolDetails"]["libraryVendorId"]

    @property
    def is_worm_storage_lock_enabled(self):
        """Treats is worm enabled as a read only attribute"""
        return self._storage_pool_properties["storagePoolDetails"]["isWormStorage"]

    @property
    def is_object_level_worm_lock_enabled(self):
        """Treats is object WORM enabled as a read only attribute"""
        worm_flag = int(self._storage_pool_properties["storagePoolDetails"]["copyInfo"]["wormStorageFlag"])
        return worm_flag & WORMLockType.OBJECT == WORMLockType.OBJECT

    @property
    def is_bucket_level_worm_lock_enabled(self):
        """Treats is bucket WORM enabled as a read only attribute"""
        worm_flag = int(self._storage_pool_properties["storagePoolDetails"]["copyInfo"]["wormStorageFlag"])
        return worm_flag & WORMLockType.BUCKET == WORMLockType.BUCKET

    @property
    def is_compliance_lock_enabled(self):
        """Treats is compliance lock enabled as a read only attribute"""
        return self._storage_pool_properties["storagePoolDetails"]["copyInfo"]["copyFlags"]["wormCopy"] == 1

    def get_copy(self):
        """ Returns the StoragePolicyCopy object of Storage Pool copy"""
        return StoragePolicyCopy(self._commcell_object, self.storage_pool_name, self.copy_name)

    def enable_compliance_lock(self):
        """ Enables compliance lock on Storage Pool Copy """
        self.get_copy().enable_compliance_lock()
        self.refresh()

    def enable_worm_storage_lock(self, retain_days):
        """
        Enable storage WORM lock on storage pool

        Args:
            retain_days    (int)   -- number of days of retention on WORM copy.

        Raises:
            SDKException:
                if response is not success.

                if reponse is empty.
        """

        request_json = {
            "storagePolicyCopyInfo": {
                "copyFlags": {
                    "wormCopy": 1
                },
                "retentionRules": {
                    "retainBackupDataForDays": retain_days
                }
            },
            "isWormStorage": True,
            "forceCopyToFollowPoolRetention": True
        }

        _STORAGE_POOL_COPY = self._commcell_object._services['STORAGE_POLICY_COPY'] % (
            self._storage_pool_id, str(self.copy_id))
        flag, response = self._commcell_object._cvpysdk_object.make_request('PUT', _STORAGE_POOL_COPY, request_json)

        if flag:
            if response.json():
                response = response.json()
                if "error" in response and response.get("error", {}).get("errorCode") != 0:
                    error_message = response.get("error", {}).get("errorMessage")
                    raise SDKException('Response', '102', error_message)
                else:
                    self.refresh()
            else:
                raise SDKException('Response', '101')
        else:
            response_string = self._commcell_object._update_response_(response.text)
            raise SDKException('Response', '101', response_string)

    def hyperscale_add_nodes(self, media_agents):
        """
        Add 3 new nodes to an existing storage pool

        args:
            media_agents      (List)   -- List of 3 media agents with name's(str)
                                            or instance of media agent's(object)

            Example: ["ma1","ma2","ma3"]

        Raises:
                SDKException:
                    if add nodes to an existing storage pool fails
        """
        if not isinstance(media_agents, list):
            raise SDKException('Storage', '101')

        mediagent_obj = []
        for media_agent in media_agents:
            if isinstance(media_agent, MediaAgent):
                mediagent_obj.append(media_agent)
            elif isinstance(media_agent, str):
                mediagent_obj.append(self._commcell_object.media_agents.get(media_agent))
            else:
                raise SDKException('Storage', '103')

        if len(mediagent_obj) <= 2:
            raise SDKException('Storage', '102', "Minimum 3 MediaAgents required")

        request_json = {
            "scaleoutOperationType": 2,
            "StoragePolicy": {
                "storagePolicyName": "{0}".format(self.storage_pool_name),
            },
            "storage": [
                {
                    "mediaAgent": {
                        "displayName": "{0}".format(mediagent_obj[0].media_agent_id),
                        "mediaAgentName": "{0}".format(mediagent_obj[0].media_agent_name)
                    }
                },
                {
                    "mediaAgent": {
                        "displayName": "{0}".format(mediagent_obj[1].media_agent_id),
                        "mediaAgentName": "{0}".format(mediagent_obj[1].media_agent_name)
                    }
                },
                {
                    "mediaAgent": {
                        "displayName": "{0}".format(mediagent_obj[2].media_agent_id),
                        "mediaAgentName": "{0}".format(mediagent_obj[2].media_agent_name)
                    }
                }
            ],
            "scaleoutConfiguration": {
                "configurationType": 1
            }
        }

        self._edit_storage_pool_api = self._commcell_object._services[
            'EDIT_STORAGE_POOL']

        flag, response = self._commcell_object._cvpysdk_object.make_request(
            'POST', self._edit_storage_pool_api, request_json
        )

        if flag:
            if response.json():
                error_code = response.json()['errorCode']

                if int(error_code) != 0:
                    error_message = response.json()['errorMessage']
                    o_str = 'Failed to add nodes to storage pool\nError: "{0}"'

                    raise SDKException('StoragePool', '102', o_str.format(error_message))
            else:
                raise SDKException('Response', '102')
        else:
            response_string = self._commcell_object._update_response_(response.text)
            raise SDKException('Response', '101', response_string)

        self.refresh()

    def hyperscale_reconfigure_storage_pool(self, storage_pool_name):
        """
        Reconfigures storage pool, for any failure during creation and expansion

        args:
          storage_pool_name (string) -- Name of the storage pools to reconfigure
        Raises:
                SDKException:
                    if reconfigure fails
        """
        if not isinstance(storage_pool_name, str):
            raise SDKException('Storage', '101')

        request_json = {

            "scaleoutOperationType": 4,
            "StoragePolicy":
                {
                    "storagePolicyName": "{0}".format(storage_pool_name),
                    "storagePolicyId": int("{0}".format(self.storage_pool_id))

                }
        }

        self._edit_storage_pool_api = self._commcell_object._services[
            'EDIT_STORAGE_POOL']

        flag, response = self._commcell_object._cvpysdk_object.make_request(
            'POST', self._edit_storage_pool_api, request_json
        )

        if flag:
            if response.json():
                error_code = response.json()['errorCode']

                if int(error_code) != 0:
                    error_message = response.json()['errorMessage']
                    o_str = 'Failed to reconfigure storage pool\nError: "{0}"'

                    raise SDKException('StoragePool', '102', o_str.format(error_message))
            else:
                raise SDKException('Response', '102')
        else:
            response_string = self._commcell_object._update_response_(response.text)
            raise SDKException('Response', '101', response_string)

        self.refresh()

    def hyperscale_replace_disk(self, disk_id, media_agent, storage_pool_name):
        """
              Replace disk action, over a media agent which is part of storage pool
               args:
                    disk_id (int) --> disk id for the disk to replace
                    media_agent (string/object) --> media agent name/ object
                    storage_pool_name (string) --> Name of the storage pools for replacemnet of disk
               Raises:
                       SDKException:
                           if replace fails
               """
        if isinstance(disk_id, str):
            disk_id = int(disk_id)
        elif not isinstance(disk_id, int):
            raise SDKException('Storage', '101')

        media_agent_obj = None
        if isinstance(media_agent, str):
            media_agent_obj = self._commcell_object.media_agents.get(media_agent)
        elif isinstance(media_agent, MediaAgent):
            media_agent_obj = media_agent
        else:
            raise SDKException('Storage', '103')

        if not isinstance(storage_pool_name, str):
            raise SDKException('Storage', '101')

        request_json = {

            "driveId": int("{0}".format(disk_id)),
            "operationType": 1,
            "mediaAgent": {
                "_type_": 11,
                "mediaAgentId": int("{0}".format(media_agent_obj.media_agent_id)),
                "mediaAgentName": "{0}".format(media_agent_obj.media_agent_name)
            },
            "scaleoutStoragePool": {
                "_type_": 160,
                "storagePoolId": int("{0}".format(self.storage_pool_id)),
                "storagePoolName": "{0}".format(self.storage_pool_name)
            }
        }

        self._replace_disk_storage_pool_api = self._commcell_object._services[
            'REPLACE_DISK_STORAGE_POOL']

        flag, response = self._commcell_object._cvpysdk_object.make_request(
            'POST', self._replace_disk_storage_pool_api, request_json
        )

        if flag:
            if response.json():
                error_code = response.json()['errorCode']

                if int(error_code) != 0:
                    error_message = response.json()['errorMessage']
                    o_str = 'Failed to replace disk\nError: "{0}"'

                    raise SDKException('StoragePool', '102', o_str.format(error_message))
            else:
                raise SDKException('Response', '102')
        else:
            response_string = self._commcell_object._update_response_(response.text)
            raise SDKException('Response', '101', response_string)

        self.refresh()

    def refresh(self):
        """Refreshes propery of the class object"""
        self._get_storage_pool_properties()

    def update_security_associations(self, associations_list, isUser=True, request_type=None, externalGroup=False):
        """
        Adds the security association on the storage pool object

        Args:
            associations_list   (list)  --  list of users to be associated
                Example:
                    associations_list = [
                        {
                            'user_name': user1,
                            'role_name': role1
                        },
                        {
                            'user_name': user2,
                            'role_name': role2
                        }
                    ]

            isUser (bool)           --    True or False. set isUser = False, If associations_list made up of user groups
            request_type (str)      --    eg : 'OVERWRITE' or 'UPDATE' or 'DELETE', Default will be OVERWRITE operation
            externalGroup (bool)    --    True or False, set externalGroup = True. If Security associations is being done on External User Groups

        Raises:
            SDKException:
                if association is not of List type
        """
        if not isinstance(associations_list, list):
            raise SDKException('StoragePool', '101')

        SecurityAssociation(self._commcell_object, self)._add_security_association(associations_list,
                                                                                   user=isUser,
                                                                                   request_type=request_type,
                                                                                   externalGroup=externalGroup)



# -*- coding: utf-8 -*-
#
# --------------------------------------------------------------------------
# Copyright Commvault Systems, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# --------------------------------------------------------------------------
#
"""
Main file for performing Cleanroom recovery operations

RecoveryGroups:     Class for representing all the recovery groups

RecoveryGroup:      Class for a single recovery group selected, and to perform operations on that recovery group

"""
from enum import Enum

from cvpysdk.exception import SDKException
from json.decoder import JSONDecodeError


class RecoveryGroups:
    """Class representing all the cleanroom recovery groups"""

    def __init__(self, commcell_object):
        """Initialize object of the RecoveryGroups class.

            Args:
                commcell_object (object)  --  instance of the Commcell class

        """
        self._commcell_object = commcell_object
        self._cvpysdk_object = commcell_object._cvpysdk_object

        self._RECOVERY_GROUPS_URL = commcell_object._services['ALL_RECOVERY_GROUPS']
        self._recovery_groups = None

        self.refresh()

    @property
    def all_groups(self):
        """Returns dict of all recovery groups.

         Returns dict    -   consists of all groups

                {
                     "group1_name": group1_id,
                     "group2_name": group2_id
                }

        """
        return self._recovery_groups

    def has_recovery_group(self, recovery_group_name):
        """Checks if a recovery group is present in the commcell.

            Args:
                recovery_group_name (str)  --  name of the recovery group

            Returns:
                bool - boolean output whether the group is present in commcell or not

            Raises:
                SDKException:
                    if type of the group name argument is not string

        """
        if not isinstance(recovery_group_name, str):
            raise SDKException('RecoveryGroup', '101')

        return self._recovery_groups and recovery_group_name in self._recovery_groups

    def get(self, recovery_group_name):
        """Returns a recovery group object.

            Args:
                recovery_group_name (str)  --  name of the recovery group

            Returns:
                object - instance of the recovery group class for the given group name

            Raises:
                SDKException:
                    if type of the group name argument is not string

                    if no group exists with the given name

        """
        if not isinstance(recovery_group_name, str):
            raise SDKException('RecoveryGroup', '101')
        else:
            if self.has_recovery_group(recovery_group_name):
                return RecoveryGroup(
                    self._commcell_object,
                    recovery_group_name,
                    self.all_groups[recovery_group_name]
                )

            raise SDKException('RecoveryGroup', '102',
                               'No recovery group exists with name: {0}'.format(recovery_group_name))

    def refresh(self):
        """Refresh the recovery groups"""
        self._recovery_groups = self._get_recovery_groups()

    def __str__(self):
        """Representation string consisting of all recovery groups .

            Returns:
                str     -   string of all the groups

        """
        representation_string = '{:^5}\t{:^20}\n\n'.format('S. No.', 'RecoveryGroup')

        for index, group in enumerate(self._recovery_groups):
            sub_str = '{:^5}\t{:20}\n'.format(
                index + 1,
                group
            )
            representation_string += sub_str

        return representation_string.strip()

    def _get_recovery_groups(self):
        """Gets all the recovery groups.

            Returns:
                dict - consists of all recovery groups in the client
                    {
                         "group1_name": group1_id,
                         "group2_name": group2_id
                    }

            Raises:
                SDKException:
                    if response is empty

                    if response is not success

        """
        flag, response = self._cvpysdk_object.make_request('GET', self._RECOVERY_GROUPS_URL)

        if flag:
            try:
                json_resp = response.json()

                group_name_id_dict = {group['name']: group['id'] for group in json_resp['recoveryGroups']}

                return group_name_id_dict
            except (JSONDecodeError, KeyError):
                raise SDKException('Response', '102')
        else:
            raise SDKException('Response', '101', self._commcell_object._update_response_(response.text))


class RecoveryStatus(Enum):
    NO_STATUS = 0
    NONE = 0
    NOT_READY = 1
    READY = 2
    RECOVERED = 3
    FAILED = 4
    RECOVERED_WITH_ERRORS = 5
    IN_PROGRESS = 6
    CLEANED_UP = 7


class RecoveryGroup:
    """Class to perform actions on a recovery group"""

    def __init__(self, commcell_object, recovery_group_name, recovery_group_id=None):
        """Initialize the instance of the RecoveryGroup class.

            Args:
                commcell_object   (object)    --  instance of the Commcell class

                recovery_group_name      (str)       --  name of the target

                recovery_group_id        (str)       --  id of the target (default: None)

        """
        self._commcell_object = commcell_object
        self._cvpysdk_object = commcell_object._cvpysdk_object

        if recovery_group_id is not None:
            self._recovery_group_id = str(recovery_group_id)
        else:
            # get id from RecoveryGroups class
            self._recovery_group_id = RecoveryGroups(commcell_object).all_groups[recovery_group_name]

        self._RECOVERY_GROUP_URL = commcell_object._services['RECOVERY_GROUP'] % self._recovery_group_id
        self._RECOVER_URL = commcell_object._services['RECOVERY_GROUP_RECOVER'] % self._recovery_group_id

        # will be set when refresh is called
        self._properties = {}

        self.refresh()

    @property
    def id(self):
        """recovery group id"""
        return int(self._recovery_group_id)

    @property
    def entities(self):
        """list of entities in recovery group"""
        return self._properties['entities']

    def _recover_entities(self, entity_ids):
        """
        Sends request to recover all entities with specified ids

        Args:
            entity_ids: iterable of entity ids

        Returns:
            job_id: job id of recovery

        Raises:
            SDKException:
                if response is empty

                if response is not success

        """
        flag, response = self._cvpysdk_object.make_request('POST', self._RECOVER_URL, payload={
            'recoveryGroup': {
                'id': self.id
            },
            'entities': [{'id': e_id} for e_id in entity_ids]
        })

        if flag:
            try:
                return response.json()['jobId']
            except (JSONDecodeError, KeyError):
                raise SDKException('Response', '102', 'Job id not found in response')
        else:
            raise SDKException('Response', '101', self._commcell_object._update_response_(response.text))

    def recover_all(self):
        """
        Sends request to recover all entities

        Returns:
            job_id: job id of recovery

        Raises:
            SDKException:
                if response is empty

                if response is not success

        """
        eligible_entities = [entity['id'] for entity in self.entities if
                             entity['recoveryStatus'] not in [RecoveryStatus.NOT_READY.value,
                                                              RecoveryStatus.IN_PROGRESS.value]]
        return self._recover_entities(eligible_entities)

    def refresh(self):
        """Refresh the recovery group"""
        self._properties = self._get_recovery_group_properties()

    def _get_recovery_group_properties(self):
        """Gets recovery group properties.

            Returns:
                dict - properties for the recovey group

            Raises:
                SDKException:
                    if response is empty

                    if response is not success

        """
        flag, response = self._cvpysdk_object.make_request('GET', self._RECOVERY_GROUP_URL)

        if flag:
            try:
                return response.json()
            except JSONDecodeError:
                raise SDKException('Response', '102')
        else:
            raise SDKException('Response', '101', self._commcell_object._update_response_(response.text))

    def delete(self):
        """
        Sends a request to delete a replication Group

        Raises:
            SDKException:
                if response is empty

                if response is not success
        """

        flag, response = self._cvpysdk_object.make_request('DELETE', self._RECOVERY_GROUP_URL)

        if flag:
            try:
                return response.json()
            except JSONDecodeError:
                raise SDKException('Response', '102')
        else:
            raise SDKException('Response', '101', self._commcell_object._update_response_(response.text))



# -*- coding: utf-8 -*-

# --------------------------------------------------------------------------
# Copyright Commvault Systems, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# --------------------------------------------------------------------------

"""File for operating on a Virtual Server Azure Instance.

AzureInstance is the only class defined in this file.

AzureInstance: Derived class from VirtualServer  Base class, representing a
                           Azure instance, and to perform operations on that instance

AzureInstance:

        _init_(self, agent, name, iid)      -- initialize object of azure Instance
                                        object associated with the VirtualServer Instance

        _get_instance_properties()          --  VirtualServer Instance class method
                                            overwritten to get Azure classic
                                            Specific instance properties as well


"""

from ..vsinstance import VirtualServerInstance
from ...exception import SDKException
from ...instance import Instance

class AzureInstance(VirtualServerInstance):

    def __init__(self, agent, name, iid):
        """Initialize the Instance object for the given Virtual Server instance.

            Args:
                class_object (agent_object,instance_name,instance_id)  --  instance of the
                                                                                Agent class,
                                                                                instance name,
                                                                                instance id

        """

        super(AzureInstance, self).__init__(agent, name, iid)
        self._vendor_id = 5

    def  _get_instance_properties(self):
        """
        Get the properties of this instance

        Raise:
            SDK Exception:
                if response is not empty
                if response is not success
        """

        super(AzureInstance, self)._get_instance_properties()
        self._server_name = []

        if 'virtualServerInstance' in self._properties:
            _member_servers = self._properties["virtualServerInstance"] \
                                                ["associatedClients"]["memberServers"]
            for _each_client in _member_servers:
                client = _each_client['client']
                if 'clientName' in client.keys():
                    self._server_name.append(str(client['clientName']))

        # waiting for praveen form

    def _get_instance_properties_json(self):
        """get the all instance related properties of this subclient.

          Returns:
               dict - all instance properties put inside a dict

        """
        instance_json = {
            "instanceProperties":{
                "isDeleted": False,
                "instance": self._instance,
                "instanceActivityControl": self._instanceActivityControl,
                "virtualServerInstance": {
                    "vsInstanceType": self._virtualserverinstance['vsInstanceType'],
                    "associatedClients": self._virtualserverinstance['associatedClients'],
                    "vmwareVendor": {}
                    }
                       }
               }
        return instance_json

    @property
    def server_name(self):
        """getter for the domain name in the Hyper-V json"""
        return self._server_name

    @property
    def server_host_name(self):
        """getter for the domain name in the vmware vendor json"""
        return self._server_name
    #return self._server_name
    # TODO will change with Praveen Form(jmalik)



# -*- coding: utf-8 -*-
# pylint: disable=R1705, R0205
# --------------------------------------------------------------------------
# Copyright Commvault Systems, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# --------------------------------------------------------------------------

"""
File for performing Ad agent dashboard operation.

AdDashboard class is defined in this file.

AdDashboard:    Class for performing Ad agent dashboard operation.

Class:
    AdDashboard :

        __init__(self,commcell_object)         --    initialize object of AdDashboard class associated with the commcell.

        get_ad_dashboard_details               --    get the dashboard details from Dashboard API.

        get_ad_apps_details                    --    get the app listing details from App Listing API.

        configured                          --    return whether both AD and Azure AD are configured or not from Dashboard API and App Listing API.

        _get_domains_and_tenants                    --    return number of domains and tenants from Dashboard API and App Listing API.

        _get_backup_health                          --    return backup health panel details from Dashboard API and App Listing API.

        _get_data_distribution                      --    return data distribution panel details from Dashboard API and App Listing API.

        _get_application_panel                      --    return application panel details from Dashboard API and App Listing API.

AdDashboard Attributes
----------------------
    **configure_dict**                          --      returns a dictionary indicating whether AD and Azure AD are configured with the commcell.

    **domains_and_tenants_dict**                --      returns a dictionary containing the number of domain controllers and tenants in the commcell.

    **backup_health_dict**                      --      returns a dictionary with information about the backup health panel of the AD Dashboard, including SLA met and SLA not met.

    **data_distribution_dict**                  --      returns a dictionary with information about the data distribution panel of the AD Dashboard, such as backup size and backed-up objects.

    **application_panel_dict**                  --      returns a dictionary with information about the application panel of the AD Dashboard, including Azure AD backup size and AD backup size.
"""

from cvpysdk.exception import SDKException

class AdDashboard(object):
    """
    Class for AD Dashboard Details
    """

    def __init__(self,commcell_object):
        """Initialize object of the Clients class.

            Args:
                commcell_object (object)  --  instance of the Commcell class

            Returns:
                object - instance of the Clients class
        """
        self._commcell_object = commcell_object
        self._cvpysdk_object = commcell_object._cvpysdk_object
        self._services = commcell_object._services
        self.dashboard_response = None
        self.azure_ad_response = None
        self.apps_response = None
        self.apps_totalentities = None

    def get_ad_dashboard_details(self):
        """
        REST API call to get AD Dashboard details in the commcell
        Raises:
            SDKException:

                if response is empty

                if response is not success
        """
        configured = self._services['ADDASHBOARD'] + '?slaNumberOfDays=1'
        flag, response = self._cvpysdk_object.make_request(method='GET', url=configured)
        if flag and response:
            self.dashboard_response = response.json()
        elif not flag:
            raise SDKException('Response', '101', self._commcell_object._update_response_(response.text))
        else:
            raise SDKException('Response', '102', self._commcell_object._update_response_(response.text))

    def get_ad_apps_details(self):
        """
        REST API call to get AD APP Listing details in the commcell
        Raises:
            SDKException:

                if response is empty

                if response is not success
        """
        configured = self._services['ADAPPS']
        flag, response = self._cvpysdk_object.make_request(method='GET', url=configured)

        if flag and response:
            self.apps_response = response.json()
        elif not flag:
            raise SDKException('Response', '101', self._commcell_object._update_response_(response.text))
        else:
            raise SDKException('Response', '102', self._commcell_object._update_response_(response.text))

    def _configured(self):
        """
        Function check both AD and Azure AD are configured
        Returns:
            configure_dict    --  Configuration value of AD and Azure AD from Dashboard API and Apps Listing API as dict
        """
        configure_dict = {"adconfigure": self.dashboard_response.get('agentSummary', [{}])[0].get('isConfigured', None),
                          "aadconfigure": self.dashboard_response.get('agentSummary', [{}])[1].get('isConfigured',None),
                          "apps_adconfigure": False,
                          "apps_aadconfigure": False}
        for i in range(self.apps_response.get('totalADClients', None)):
            if self.apps_response.get('adClients', [{}])[i].get('isConfigured') and \
                    self.apps_response.get('adClients', [{}])[i].get('appTypeId') == 41:
                configure_dict["apps_adconfigure"] = True
                break

        for i in range(self.apps_response.get('totalADClients', None)):
            if self.apps_response.get('adClients', [{}])[i].get('isConfigured') and \
                    self.apps_response.get('adClients', [{}])[i].get('appTypeId') == 139:
                configure_dict["apps_aadconfigure"] = True
                break

        return configure_dict

    @property
    def is_configured(self):
        """
        Returns:
            configure_dict    --  Configuration value of AD and Azure AD from Dashboard API and Apps Listing API as dict
        """
        configure_dict=self._configured()
        return configure_dict

    def _get_domains_and_tenants(self):
        """
        Function to get number of domains and tenants
        Returns:
            domains_and_tenants_dict    --  Number of domain controllers and tenants from Dashboard API and Apps Listing API as dict
        """
        domains_and_tenants_dict = {
            "total_entities": self.dashboard_response.get('solutionSummary', {}).get('slaSummary', {}).get(
                'totalEntities'),
            "domain_controllers": self.dashboard_response.get('agentSummary', [{}])[0].get('slaSummary', {}).get(
                'totalEntities'),
            "tenants": self.dashboard_response.get('agentSummary', [{}])[1].get('slaSummary', {}).get('totalEntities'),
            "apps_totalentities": 0,
            "apps_domain_controllers": 0,
            "apps_tenants": 0
            }
        for i in range(len(self.apps_response['adClients'])):
            if self.apps_response.get('adClients', [])[i].get('appTypeId') == 41 and \
                    self.apps_response.get('adClients', [])[i].get('slaStatus') != "EXCLUDED_SLA":
                domains_and_tenants_dict["apps_domain_controllers"] += 1
            if self.apps_response.get('adClients', [])[i].get('appTypeId') == 139 and \
                    self.apps_response.get('adClients', [])[i].get('slaStatus') != "EXCLUDED_SLA":
                domains_and_tenants_dict["apps_tenants"] += 1
            if self.apps_response.get('adClients', [])[i].get('slaStatus') != "EXCLUDED_SLA":
                domains_and_tenants_dict["apps_totalentities"] += 1
        return domains_and_tenants_dict

    @property
    def domains_and_tenants(self):
        """
        Returns:
            domains_and_tenants_dict    --  Number of domain controllers and tenants from Dashboard API and Apps Listing API as dict
        """
        return self._get_domains_and_tenants()

    def _get_backup_health(self):
        """
        Function to get backup health panel details
        Returns:
            backup_health_dict    --  Backup health panel details as dict
        """
        backup_health_dict = {
            "recently_backedup": self.dashboard_response.get('solutionSummary', {}).get('slaSummary', {}).get(
                'totalEntities', 0) - self.dashboard_response.get('solutionSummary', {}).get('slaSummary', {}).get(
                'slaNotMetEntities', 0),
            "recently_backedup_per": self.dashboard_response.get('solutionSummary', {}).get('slaSummary', {}).get(
                'slaMetPercentage', 0),
            "recently_not_backedup": self.dashboard_response.get('solutionSummary', {}).get('slaSummary', {}).get(
                'slaNotMetEntities', 0) - self.dashboard_response.get('solutionSummary', {}).get('slaSummary', {}).get(
                'neverBackedupEntities', 0),
            "recently_not_backedup_per": self.dashboard_response.get('solutionSummary', {}).get('slaSummary', {}).get(
                'slaNotMetProcessedAtleastOncePercentage', 0),
            "never_backedup": self.dashboard_response.get('solutionSummary', {}).get('slaSummary', {}).get(
                'neverBackedupEntities', 0),
            "never_backedup_per": self.dashboard_response.get('solutionSummary', {}).get('slaSummary', {}).get(
                'neverBackedupPercentage', 0),
            "apps_recently_backedup": 0,
            "apps_recently_not_backedup": 0,
            "apps_never_backedup": 0,
            "apps_totalentities": 0,
            "apps_recently_backedup_per": 0,
            "apps_recently_not_backedup_per": 0,
            "apps_never_backedup_per": 0
            }
        for i in range(len(self.apps_response['adClients'])):
            if self.apps_response.get('adClients', [])[i].get('slaStatus') == "MET_SLA":
                backup_health_dict["apps_recently_backedup"] += 1
            if self.apps_response.get('adClients', [])[i].get('slaStatus') == "MISSED_SLA" and \
                    self.apps_response.get('adClients', [])[i].get('numberOfItems') != 0:
                backup_health_dict["apps_recently_not_backedup"] += 1
            if self.apps_response.get('adClients', [])[i].get('slaStatus') == "MISSED_SLA" and \
                    self.apps_response.get('adClients', [])[i].get('numberOfItems') == 0:
                backup_health_dict["apps_never_backedup"] += 1
            if self.apps_response.get('adClients', [])[i].get('slaStatus') != "EXCLUDED_SLA":
                backup_health_dict["apps_totalentities"] += 1

        backup_health_dict["apps_recently_backedup_per"] = (
            round(((backup_health_dict["apps_recently_backedup"] / backup_health_dict["apps_totalentities"]) * 100), 2))
        backup_health_dict["apps_recently_not_backedup_per"] = (
            round(((backup_health_dict["apps_recently_not_backedup"] / backup_health_dict["apps_totalentities"]) * 100),
                  2))
        backup_health_dict["apps_never_backedup_per"] = (
            round(((backup_health_dict["apps_never_backedup"] / backup_health_dict["apps_totalentities"]) * 100), 2))

        return backup_health_dict

    @property
    def backup_health(self):
        """
        Returns:
            backup_health_dict    --  Backup health panel details as dict
        """
        return self._get_backup_health()


    def _get_data_distribution(self):
        """
        Function to get data distribution panel details
        Returns:
            data_distribution_dict    --  Data distribution details data as dict
        """
        data_distribution_dict = {"backup_size": round((((self.dashboard_response.get('agentSummary', [{}])[0].get('applicationSize', 0) +
                                                          self.dashboard_response.get('agentSummary', [{}])[1].get('applicationSize',0))
                                                         / 1024) / 1024), 2),
                                  "backup_obj": (
                                              self.dashboard_response.get('agentSummary', [{}])[0].get('numberOfItems', 0) +
                                              self.dashboard_response.get('agentSummary', [{}])[1].get('numberOfItems',0)),
                                  "apps_backup_size": 0,
                                  "apps_backup_obj": 0,
                                  }

        for i in range(len(self.apps_response['adClients'])):
            data_distribution_dict["apps_backup_size"] += self.apps_response.get('adClients', [{}])[i].get('applicationSize', None)

        data_distribution_dict["apps_backup_size"] = round(((data_distribution_dict["apps_backup_size"] / 1024) / 1024), 2)

        for i in range(len(self.apps_response['adClients'])):
            data_distribution_dict["apps_backup_obj"] += self.apps_response.get('adClients', [{}])[i].get('numberOfItems', None)

        return data_distribution_dict

    @property
    def data_distribution(self):
        """
        Returns:
            data_distribution_dict    --  Data distribution details data as dict
        """
        return self._get_data_distribution()

    def _get_application_panel(self):
        """
        Function to get application panel details
        Returns:
            application_panel_dict    --  Application panel details as dict
        """
        application_panel_dict = {
            "aad_tenant": self.dashboard_response.get('agentSummary', [{}])[1].get('slaSummary', {}).get('totalEntities'),
            "aad_backup_size": round((((self.dashboard_response.get('agentSummary', [{}])[1].get('applicationSize')) / 1024) / 1024), 2),
            "aad_backup_obj": self.dashboard_response.get('agentSummary', [{}])[1].get('numberOfItems'),
            "aad_sla_per": self.dashboard_response.get('agentSummary', [{}])[1].get('slaSummary', {}).get('slaMetPercentage'),
            "aad_not_sla_per": self.dashboard_response.get('agentSummary', [{}])[1].get('slaSummary', {}).get('slaNotMetProcessedAtleastOncePercentage'),

            "ad_domains": self.dashboard_response.get('agentSummary', [{}])[0].get('slaSummary', {}).get('totalEntities'),
            "ad_backup_size": round((((self.dashboard_response.get('agentSummary', [{}])[0].get('applicationSize')) / 1024) / 1024), 2),
            "ad_backup_obj": self.dashboard_response.get('agentSummary', [{}])[0].get('numberOfItems'),
            "ad_sla_per": self.dashboard_response.get('agentSummary', [{}])[0].get('slaSummary', {}).get('slaMetPercentage'),
            "ad_not_sla_per": self.dashboard_response.get('agentSummary', [{}])[0].get('slaSummary', {}).get('slaNotMetProcessedAtleastOncePercentage'),

            "apps_aad_tenant": 0, "apps_aad_backup_size": 0, "apps_aad_backup_obj": 0,
            "apps_aad_sla_per": 0,"apps_aad_not_sla_per": 0,
            "apps_ad_domains": 0, "apps_ad_backup_size": 0, "apps_ad_backup_obj": 0,
            "apps_ad_sla_per": 0,"apps_ad_not_sla_per": 0,
            "apps_aad_sla": 0, "apps_ad_sla": 0}

        for i in range(len(self.apps_response['adClients'])):
            if self.apps_response.get('adClients', [])[i].get('appTypeId') == 139 and \
                    self.apps_response.get('adClients', [])[i].get('slaStatus') != "EXCLUDED_SLA":
                application_panel_dict["apps_aad_tenant"] += 1

            if self.apps_response.get('adClients', [])[i].get('appTypeId') == 139:
                application_panel_dict["apps_aad_backup_size"] += self.apps_response.get('adClients', [{}])[i].get(
                    'applicationSize')
                application_panel_dict["apps_aad_backup_obj"] += self.apps_response.get('adClients', [{}])[i].get(
                    'numberOfItems')

            if self.apps_response.get('adClients', [])[i].get('appTypeId') == 41 and \
                    self.apps_response.get('adClients', [])[i].get('slaStatus') != "EXCLUDED_SLA":
                application_panel_dict["apps_ad_domains"] += 1

            if self.apps_response.get('adClients', [])[i].get('appTypeId') == 41:
                application_panel_dict["apps_ad_backup_size"] += self.apps_response.get('adClients', [{}])[i].get(
                    'applicationSize')
                application_panel_dict["apps_ad_backup_obj"] += self.apps_response.get('adClients', [{}])[i].get(
                    'numberOfItems')

            if self.apps_response.get('adClients', [])[i].get('appTypeId') == 139 and \
                    self.apps_response.get('adClients', [])[i].get('slaStatus') == "MET_SLA":
                application_panel_dict["apps_aad_sla"] += 1

            if self.apps_response.get('adClients', [])[i].get('appTypeId') == 41 and \
                    self.apps_response.get('adClients', [])[i].get('slaStatus') == "MET_SLA":
                application_panel_dict["apps_ad_sla"] += 1

        application_panel_dict["apps_aad_backup_size"] = round(
            ((application_panel_dict["apps_aad_backup_size"] / 1024) / 1024), 2)
        application_panel_dict["apps_ad_backup_size"] = round(
            ((application_panel_dict["apps_ad_backup_size"] / 1024) / 1024), 2)

        application_panel_dict["apps_aad_sla_per"] = (
            round(((application_panel_dict["apps_aad_sla"] / application_panel_dict["apps_aad_tenant"]) * 100), 2))
        application_panel_dict["apps_ad_sla_per"] = (
            round(((application_panel_dict["apps_ad_sla"] / application_panel_dict["apps_ad_domains"]) * 100), 2))

        application_panel_dict["apps_aad_not_sla_per"] = (round((((application_panel_dict["apps_aad_tenant"] -
                                                                   application_panel_dict["apps_aad_sla"]) /
                                                                  application_panel_dict["apps_aad_tenant"]) * 100), 2))
        application_panel_dict["apps_ad_not_sla_per"] = (round((((application_panel_dict["apps_ad_domains"] -
                                                                  application_panel_dict["apps_ad_sla"]) /
                                                                 application_panel_dict["apps_ad_domains"]) * 100), 2))

        return application_panel_dict

    @property
    def application_panel(self):
        """
        Returns:
            application_panel_dict    --  Application panel details as dict
        """
        return self._get_application_panel()


# -*- coding: utf-8 -*-

# --------------------------------------------------------------------------
# Copyright Commvault Systems, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# --------------------------------------------------------------------------

"""Main file for performing Cleanroom Target specific operations.

CleanroomTargets and CleanroomTarget are 2 classes defined in this file.

CleanroomTargets:     Class for representing all the cleanroom targets

CleanroomTarget:      Class for a single cleanroom target selected, and to perform operations on that cleanroom target


cleanroomTargets:
    __init__()                   --  initialize object of CleanroomTargets class

    __str__()                   --  returns all the Cleanroom Targets

    _get_cleanroom_targets()     -- Gets all the cleanroom targets

    has_cleanroom_target()       -- Checks if a target is present in the commcell.

    get()                        --  returns the cleanroom target class object of the input target name

    refresh()                   --  refresh the targets present in the client

cleanroomTargets Attributes
--------------------------

    **all_targets**             --  returns the dictionary consisting of all the targets that are
    present in the commcell and their information such as id and name

CleanroomTarget:
    __init__()                   --   initialize object of CleanroomTarget with the specified cleanroom target name

    _get_cleanroom_target_id()   --   method to get the cleanroom target id

    _get_cleanroom_target_properties()  --   get the properties of this recovery target

    _delete_cleanroom_target()    -- Deletes the cleanroom target

    delete()                     -- Deletes the cleanroom target

    refresh()                   --   refresh the object properties

CleanroomTarget Attributes
--------------------------

    **cleanroom_target_id**      -- Returns the id of the cleanroom target
    **cleanroom_target_name**    -- Returns the name of the cleanroom Target
    **destination_hypervisor**  -- Returns the name of destination hypervisor
    **vm_prefix**               -- Returns the prefix of the vm name
    **destination_host**        -- Returns the destination host
    **storage_account**          -- Returns the storage_account host
    **policy_type**             -- Returns the policy type
    **application_type**          -- Returns the application type
    **restore_as_managed_vm**   -- Returns the restore_as_managed_vm
    **region**                  -- Returns the region
    **expiration_time**         -- Returns the _expiration_time
    **vm_suffix**               -- Returns the vm_suffix
    **vm_prefix**               -- Returns the vm_prefix
    **access_node**             -- Returns the access_node
    **access_node_client_group** -- Returns the access_node_client_group

"""
from __future__ import absolute_import
from __future__ import unicode_literals

from cvpysdk.exception import SDKException


class CleanroomTargets:

    """Class for representing all the Cleanroom targets"""

    def __init__(self, commcell_object):
        """Initialize object of the CleanroomTargets class.

            Args:
                commcell_object (object)  --  instance of the Commcell class

        """
        self._commcell_object = commcell_object

        self._cvpysdk_object = commcell_object._cvpysdk_object
        self._services = commcell_object._services
        self._update_response_ = commcell_object._update_response_
        self._RECOVERY_TARGETS_API = self._services['GET_ALL_RECOVERY_TARGETS']

        self._cleanroom_targets = None
        self.refresh()

    def __str__(self):
        """Representation string consisting of all targets .

            Returns:
                str     -   string of all the targets

        """
        representation_string = '{:^5}\t{:^20}\n\n'.format('S. No.', 'CleanroomTargets')

        for index, cleanroom_target in enumerate(self._cleanroom_targets):
            sub_str = '{:^5}\t{:20}\n'.format(
                index + 1,
                cleanroom_target
            )
            representation_string += sub_str

        return representation_string.strip()

    def _get_cleanroom_targets(self):
        """Gets all the cleanroom targets.

            Returns:
                dict - consists of all targets in the client
                    {
                         "target1_name": target1_id,
                         "target2_name": target2_id
                    }

            Raises:
                SDKException:
                    if response is empty

                    if response is not success

        """
        flag, response = self._cvpysdk_object.make_request('GET', self._RECOVERY_TARGETS_API)
        if flag:
            if response.json() and 'recoveryTargets' in response.json():
                cleanroom_target_dict = {}
                for cleanroomTarget in response.json()['recoveryTargets']:
                    if cleanroomTarget['applicationType'] == "CLEAN_ROOM":
                        temp_name = cleanroomTarget['name'].lower()
                        cleanroom_target_dict[temp_name] = str(cleanroomTarget['id'])

                return cleanroom_target_dict
            else:
                raise SDKException('Response', '102')
        else:
            raise SDKException('Response', '101', self._update_response_(response.text))

    @property
    def all_targets(self):
        """Returns dict of all the targets.

         Returns dict    -   consists of all targets

                {
                    "target1_name": target1_id,

                    "target2_name": target2_id
                }

        """
        return self._cleanroom_targets

    def has_cleanroom_target(self, target_name):
        """Checks if a target is present in the commcell.

            Args:
                target_name (str)  --  name of the target

            Returns:
                bool - boolean output whether the target is present in commcell or not

            Raises:
                SDKException:
                    if type of the target name argument is not string

        """
        if not isinstance(target_name, str):
            raise SDKException('Target', '101')

        return self._cleanroom_targets and target_name.lower() in self._cleanroom_targets

    def get(self, cleanroom_target_name):
        """Returns a target object.

            Args:
                cleanroom_target_name (str)  --  name of the target

            Returns:
                object - instance of the target class for the given target name

            Raises:
                SDKException:
                    if type of the target name argument is not string

                    if no target exists with the given name

        """
        if not isinstance(cleanroom_target_name, str):
            raise SDKException('Target', '101')
        else:
            cleanroom_target_name = cleanroom_target_name.lower()

            if self.has_cleanroom_target(cleanroom_target_name):
                return CleanroomTarget(
                    self._commcell_object, cleanroom_target_name, self.all_targets[cleanroom_target_name])

            raise SDKException('RecoveryTarget', '102', 'No target exists with name: {0}'.format(cleanroom_target_name))

    def refresh(self):
        """Refresh the cleanroom targets"""
        self._cleanroom_targets = self._get_cleanroom_targets()


class CleanroomTarget:
    """Class for performing target operations"""

    def __init__(self, commcell_object, cleanroom_target_name, cleanroom_target_id=None):
        """Initialize the instance of the CleanroomTarget class.

            Args:
                commcell_object   (object)    --  instance of the Commcell class

                cleanroom_target_name      (str)       --  name of the target

                cleanroom_target_id        (str)       --  id of the target

        """
        self._commcell_object = commcell_object

        self._cvpysdk_object = commcell_object._cvpysdk_object
        self._services = commcell_object._services
        self._update_response_ = commcell_object._update_response_
        self._cleanroom_target_name = cleanroom_target_name.lower()

        if cleanroom_target_id:
            # Use the target id mentioned in the arguments
            self._cleanroom_target_id = str(cleanroom_target_id)
        else:
            # Get the target id if target id is not provided
            self._cleanroom_target_id = self._get_cleanroom_target_id()
        self._RECOVERY_TARGET_API = self._services['GET_RECOVERY_TARGET'] % self._cleanroom_target_id

        self._cleanroom_target_properties = None

        self._policy_type = None
        self._application_type = None
        self._destination_hypervisor = None
        self._access_node = None
        self._access_node_client_group = None
        self._users = []
        self._user_groups = []
        self._vm_prefix = ''
        self._vm_suffix = ''
        self._expiration_time = None

        self._region = None
        self._availability_zone = None
        self._storage_account = None
        self._restore_as_managed_vm = None

        self.refresh()

    def _get_cleanroom_target_id(self):
        """Gets the target id associated with this target.

            Returns:
                str - id associated with this target

        """
        target = CleanroomTargets(self._commcell_object)
        return target.all_targets[self._cleanroom_target_name]

    def _delete_cleanroom_target(self):
        """Deletes the cleanroom target

            Raises:
                SDKException:
                    if response is not success
        """
        flag, response = self._cvpysdk_object.make_request('DELETE', self._RECOVERY_TARGET_API)
        if flag:
            return flag
        else:
            raise SDKException('Response', '101', self._update_response_(response.text))

    def _set_policy_type(self, policy_type):
        """Sets the policy type"""
        if policy_type == "AMAZON":
            self._policy_type = 1
        elif policy_type == "MICROSOFT":
            self._policy_type = 2
        elif policy_type == "AZURE_RESOURCE_MANAGER":
            self._policy_type = 7
        elif policy_type in ["VMW_BACKUP_LABTEMPLATE", "VMW_LIVEMOUNT"]:
            self._policy_type = 13
        else:
            self._policy_type = -1

    def _get_cleanroom_target_properties(self):
        """Gets the target properties of this target.

            Raises:
                SDKException:
                    if response is empty

                    if response is not success

        """
        flag, response = self._cvpysdk_object.make_request('GET', self._RECOVERY_TARGET_API)

        if flag:
            if response.json() and 'entity' in response.json():
                self._cleanroom_target_properties = response.json()
                self._application_type = self._cleanroom_target_properties['entity']['applicationType']
                self._destination_hypervisor = self._cleanroom_target_properties['entity']['destinationHypervisor'][
                    'name']
                self._vm_suffix = self._cleanroom_target_properties["vmDisplayName"].get("suffix", "")
                self._vm_prefix = self._cleanroom_target_properties["vmDisplayName"].get("prefix", "")
                self._access_node = self._cleanroom_target_properties["accessNode"].get("type", "")
                self._access_node_client_group = (self._cleanroom_target_properties.get('proxyClientGroupEntity', {})
                                                  .get('clientGroupName'))
                self._users = self._cleanroom_target_properties.get('securityOptions', {}).get('users', [])
                self._user_groups = self._cleanroom_target_properties.get('securityOptions', {}).get('userGroups', [])
                policy_type = self._cleanroom_target_properties["entity"].get("policyType", "")
                self._set_policy_type(policy_type)

                if self._policy_type == 7:
                    self._region = (self._cleanroom_target_properties.get('cloudDestinationOptions', {})
                                    .get('region', {})
                                    .get('name'))
                    self._availability_zone = (self._cleanroom_target_properties.get('cloudDestinationOptions', {})
                                               .get('availabilityZone'))
                    self._storage_account = (self._cleanroom_target_properties.get("destinationOptions", {})
                                             .get("dataStore", ""))

                    self._vm_size = (self._cleanroom_target_properties.get('amazonPolicy', {})
                                     .get('vmInstanceTypes', [{}])[0]
                                     .get('vmInstanceTypeName', ''))
                    self._disk_type = (self._cleanroom_target_properties.get('cloudDestinationOptions', {})
                                       .get('volumeType'))
                    self._virtual_network = (self._cleanroom_target_properties.get('networkOptions', {})
                                             .get('networkCard', {})
                                             .get('networkDisplayName'))
                    self._security_group = (self._cleanroom_target_properties.get('securityOptions', {})
                                            .get('securityGroups', [{}])[0]
                                            .get('name', ''))
                    self._create_public_ip = (self._cleanroom_target_properties.get('cloudDestinationOptions', {})
                                              .get('publicIP'))
                    self._restore_as_managed_vm = (self._cleanroom_target_properties.get('cloudDestinationOptions', {})
                                                   .get('restoreAsManagedVM'))
                    expiry_hours = (self._cleanroom_target_properties.get("liveMountOptions", {})
                                    .get("expirationTime", {})
                                    .get("minutesRetainUntil", ""))
                    expiry_days = (self._cleanroom_target_properties.get("liveMountOptions", {})
                                   .get("expirationTime", {})
                                   .get("daysRetainUntil", ""))
                    if expiry_hours:
                        self._expiration_time = f'{expiry_hours} hours'
                    elif expiry_days:
                        self._expiration_time = f'{expiry_days} days'
                    self._test_virtual_network = (self._cleanroom_target_properties.get('networkOptions', {})
                                                  .get('cloudNetwork', {})
                                                  .get('label'))
                    self._test_vm_size = (self._cleanroom_target_properties.get('amazonPolicy', {})
                                          .get('vmInstanceTypes', [{}])[0]
                                          .get('vmInstanceTypeName', ''))
            else:
                raise SDKException('Response', '102')
        else:
            raise SDKException('Response', '101', self._update_response_(response.text))

    @property
    def cleanroom_target_id(self):
        """Returns: (str) the id of the cleanroom target"""
        return self._cleanroom_target_id

    @property
    def cleanroom_target_name(self):
        """Returns: (str) the display name of the cleanroom target"""
        return self._cleanroom_target_name

    @property
    def policy_type(self):
        """Returns: (str) the policy type ID
            1  - AWS
            2  - Microsoft Hyper-V
            7  - Azure
            13 - VMware
        """
        return self._policy_type

    @property
    def application_type(self):
        """Returns: (str) the name of the application type
            0 - Replication type
            1 - Regular type
        """
        return self._application_type

    @property
    def destination_hypervisor(self):
        """Returns: (str) the client name of destination hypervisor"""
        return self._destination_hypervisor

    @property
    def access_node(self):
        """Returns: (str) the client name of the access node/proxy of the cleanroom target"""
        return self._access_node

    @property
    def access_node_client_group(self):
        """Returns: (str) The client group name set on the access node field of cleanroom target"""
        return self._access_node_client_group

    @property
    def security_user_names(self):
        """Returns: list<str> the names of the users who are used for ownership of the hypervisor and VMs"""
        return [user['userName'] for user in self._users]

    @property
    def vm_prefix(self):
        """Returns: (str) the prefix of the vm name to be prefixed to the destination VM"""
        return self._vm_prefix

    @property
    def vm_suffix(self):
        """Returns: (str) the suffix of the vm name to be suffixed to the destination VM"""
        return self._vm_suffix

    @property
    def expiration_time(self):
        """Returns: (str) VMware/Azure: the expiration time of the test boot VM/test failover VM
            eg: 4 hours or 3 days
        """
        return self._expiration_time

    @property
    def storage_account(self):
        """Returns: (str) Azure: the storage account name used to deploy the VM's storage"""
        return self._storage_account

    @property
    def region(self):
        """Return: (str) Azure: the cleanroom target region for destination VM"""
        return self._region

    @property
    def restore_as_managed_vm(self):
        """Returns: (bool) whether the destination VM will be a managed VM"""
        return self._restore_as_managed_vm

    def refresh(self):
        """Refresh the properties of the cleanroom Target."""
        self._get_cleanroom_target_properties()

    def delete(self):
        """Deletes the Cleanroom Target. Returns: (bool) whether the target is deleted or not."""
        return self._delete_cleanroom_target()



# -*- coding: utf-8 -*-

# --------------------------------------------------------------------------
# Copyright Commvault Systems, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# --------------------------------------------------------------------------

"""File for operating on a Virtual Server Xen Server Instance.

XenServer Instance is the only class defined in this file.

XenInstance: Derived class from VirtualServer  Base class, representing a
                           Xen Server instance, and to perform operations on that instance

XenInstance:

    __init__(agent_object,instance_name,instance_id)    -- initialize object of FusionCompute
                                                            Instance object associated with the
                                                            VirtualServer Instance


    _get_instance_properties()                          --  VirtualServer Instance class method
                                                            overwritten to get Xen Server specific
                                                            instance properties as well

    _set_instance_properties()                          --  Xen Server Instance class method
                                                            to set Xen Specific instance properties


"""

from ..vsinstance import VirtualServerInstance


class Xen(VirtualServerInstance):
    """Class for representing an Xen Server instance of the Virtual Server agent."""

    def __init__(self, agent, instance_name, instance_id=None):
        """Initialize the Instance object for the given Virtual Server instance.

            Args:
                class_object (agent_object,instance_name,instance_id)  --  instance of the
                                                                            Agent class,
                                                                            instance name,
                                                                            instance id

        """
        super(Xen, self).__init__(agent, instance_name, instance_id)
        self._vendor_id = 3
        self._server_name = None
        self._server_host_name = None

    def _get_instance_properties(self):
        """
        Get the properties of this instance

        Raise:
            SDK Exception:
                if response is not empty
                if response is not success
        """

        super(Xen, self)._get_instance_properties()
        self._server_name = self._instance.get('clientName', '')

    def _get_instance_properties_json(self):
        """get the all instance related properties of this subclient.

          Returns:
               dict - all instance properties put inside a dict

        """
        instance_json = {
            "instanceProperties": {
                "isDeleted": False,
                "instance": self._instance,
                "instanceActivityControl": self._instanceActivityControl,
                "virtualServerInstance": {
                    "vsInstanceType": self._virtualserverinstance['vsInstanceType'],
                    "associatedClients": self._virtualserverinstance['associatedClients'],
                    "vmwareVendor": {}
                }
            }
        }
        return instance_json

    @property
    def server_host_name(self):
        """Getter for server_host_name property"""
        # This property will be set during TC execution. 
        return self._server_host_name

    @server_host_name.setter
    def server_host_name(self, value):
        """Setter for server_host_name property"""
        self._server_host_name = value

    @property
    def server_name(self):
        """Getter for the server_name property"""
        return self._server_name
